/* 
 * Copyright (C) 1999-2004 Joachim Wieland <joe@mcknight.de>
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place - Suite 330, Boston, MA 02111, USA.
 */

#include "ftp_main.h"

static struct portrangestruct *actvrangeserver = (struct portrangestruct*) 0;
static struct portrangestruct *actvrangeclient = (struct portrangestruct*) 0;
static void sayact(int, char*, int);

int portcommandcheck(const char* buffer, struct sockaddr_in* sin,
			struct proxyinfo* pi, struct messagebuffer* msgbuf) {
	int cs = pi->si->connection.clientsocket;
	int ret;

	ret = parsesock(buffer, sin, ACTIVE);
	if (ret) {
		jlog(3, "Error parsing the PORT command: %s", buffer);
		message_buffer_add(msgbuf, cs,
				501, "Illegal PORT command\r\n");
		return -1;
	}
	if (config_get_bool("allowforeignaddress", 0) == 0) {
		char* sin_addr = strdup(inet_ntoa(sin->sin_addr));
		if (strcmp(sin_addr, get_char_ip(GET_IP_CLIENT, pi)) != 0) {
			jlog(3, "Illegal foreign address %s in PORT command.",
								sin_addr);
			jlog(5, "Expected address %s instead",
					get_char_ip(GET_IP_CLIENT, pi));
			message_buffer_add(msgbuf, cs,
				500, "Illegal address in PORT command\r\n");
			free(sin_addr);
			return -1;
		} else {
			free(sin_addr);
		}
	}
	if (config_get_bool("allowreservedports", 0) == 0) {
		/* XXX verify */
		if (ntohs(sin->sin_port) < IPPORT_RESERVED) {
			jlog(3, "Illegal port number %d in PORT command.",
						ntohs(sin->sin_port));
			message_buffer_add(msgbuf, cs,
				500, "Illegal port number in PORT command\r\n");
			return -1;
		}
	}
	return 0;
}


/* 
 * activeclient() reads the "PORT x,x,x,x,x,x" command from the client,
 * opens that port on the client's side and returns the socket descriptor
 * 
 * */

int activeclient(char* buffer, struct proxyinfo* pi,
					struct messagebuffer* msgbuf) {
	struct sockaddr_in sin;
	struct portrangestruct* prs;
	struct clientinfo* ci = pi->ci;
	int ret;

	/* returns 0 if ok, -1 in case of an error */
	ret = portcommandcheck(buffer, &sin, pi, msgbuf);
	if (ret < 0) {
		return CMD_ERROR;
	}
	jlog(8, "Opening the active FTP port %d on %s",
			ntohs(sin.sin_port), inet_ntoa(sin.sin_addr));
	if (ci->data_connection.waitforconnect
				== &(ci->data_connection.dataclientsock)) {

		ci->data_connection.waitforconnect = (int*) 0;
	}
	if (!actvrangeclient) {
		const char* range = config_get_option("activeportrangeclient");
		if (!range) {
			range = config_get_option("activeportrange");
		}
		actvrangeclient = config_parse_portranges(range);
	}
	if (!actvrangeclient) {
		prs = portrangestruct_port2portrange(ci->data_connection.dataport);
	} else {
		prs = actvrangeclient;
	}
	ret = openportiaddr(sin.sin_addr.s_addr,          /* dest ip   */
			    ntohs(sin.sin_port),          /* dest port */
			    /* source ip, port */
			    /* this variable is set after a successul login and
			     * respects the dataclientaddress option */
			    ci->data_connection.data_addr_to_client,
			    prs, pi);
	if (ret < 0) {
		jlog(8, "setting dataclientsock to -1 (openport error)");
		ci->data_connection.dataclientsock = -1;
		return CMD_ERROR;
	}
	ci->data_connection.dataclientsock = ret;
	/* XXX free prs
	 *
	 * attention: port2portrange returns a static struct !
	 * 
	 * ? */

	return CMD_HANDLED;
}


/*
 * activeserver() opens a port on the proxy machine, determines the port
 * number and tells it to the server (along with the address of the interface
 * over which we connect to it).
 *
 * Returns:
 * 	TRANSFERMODE_OK
 * 	TRANSFERMDOE_FAIL
 * 	TRANSFERMODE_FATAL
 *
 */

int activeserver(struct proxyinfo *pi, struct messagebuffer* msgbuf,
					struct ftp_message* server_msg) {
	int ret;
	struct sockaddr_in sin;
	int ss = pi->si->connection.serversocket;
	int cs = pi->si->connection.clientsocket;
	struct clientinfo* ci = pi->ci;
	struct serverinfo* si = pi->si;
	const char* opt;
	struct ip_t display_ip;
	char* display_char;

	if (!actvrangeserver) {
		const char* range = config_get_option("activeportrangeserver");
		if (!range) {
			range = config_get_option("activeportrange");
		}
		/* XXX free this ? */
		actvrangeserver = config_parse_portranges(range);
	}
	/* open a port on our side use the interface address through which
	 * we are connected to the server */
	ret = openlocalport(&sin,    /* for the return value */
			    /* source ip */
			    ci->data_connection.data_addr_to_server,
			    /* source ports */
			    actvrangeserver,
			    pi);
	if (ci->data_connection.waitforconnect
			== &(ci->data_connection.dataserversock)) {
		ci->data_connection.waitforconnect = (int*) 0;
	}
	if (ret < 0) {
		message_buffer_add(msgbuf, cs,
			425, "Could not establish a connection endpoint\r\n");
		return TRANSFERMODE_FAIL;
	}

	ci->data_connection.dataserversock = ret;
	ci->data_connection.waitforconnect
			= &ci->data_connection.dataserversock;

	/* tell the server about the addr + port it can connect to */
	if ((opt = config_get_option("displayserveraddress"))) {
		display_ip = ip_parse_from_char(opt);
		if (display_ip.error) {
			jlog(5, "Could not parse IP from option "
					"displayserveraddress: %s", opt);
		}
	} else if (sin.sin_addr.s_addr == INADDR_ANY) {
		/* the user can specify 0.0.0.0 with the dataserveraddress
		 * parameter. In this case, we do not want to display
		 * 0,0,0,0,x,y in a response... */
		display_ip = ip_from_s_addr(si->connection.addr_to_client);
	} else {
		display_ip = ip_from_s_addr(sin.sin_addr.s_addr);
	}
	display_char = ip_to_mchar(display_ip);
	sayact(ss, display_char, ntohs(sin.sin_port));
	free(display_char);

	*server_msg = ftp_readall(ss);
	if (server_msg->error) {
		/* Treat the timeout with a special message, but default
		 * behaviour for the other cases */
		if (server_msg->error_type == ERROR_TIMEOUT) {
			jlog(5, "Timed out reading answer for the PORT command"
					" from fd %d", server_msg->from_fd);
			message_buffer_add(msgbuf, cs,
					500, "Timeout waiting for an answer "
						"for the PORT command\r\n");
			return TRANSFERMODE_FATAL;
		} else if (server_msg->error_type == ERROR_CONNECTIONCLOSED) {
			message_buffer_add(msgbuf, cs,
					500, "Remote server closed the "
						"connection\r\n");
			return TRANSFERMODE_FATAL;
		} else {
			ftp_message_say_error(*server_msg, msgbuf, cs,
							__FILE__, __LINE__);
			return TRANSFERMODE_FAIL;
		}
	}
	/* code 200 would be "PORT command successful" */
	if (server_msg->code != 200) {
		jlog(5, "Failure sending PORT command to the server: %s",
							server_msg->msg);
		message_buffer_add(msgbuf, cs,
				425, "Can't establish connection\r\n");
		return TRANSFERMODE_FAIL;
	}
	return TRANSFERMODE_OK;
}

/* sayact() transforms the parameters ADDR and PORT to the 
 * "PORT x,x,x,x,x,x" notation and says it to the descriptor fd
 *
 * Paramters: fd: the file descriptor the command is written to
 *            addr: the address that should be included in dot notation
 *                  format
 *            port: the port that was opened on addr's side
 */

static void sayact(int fd, char* addr, int port) {
	unsigned short int lo, hi;
	char buffer[MAX_LINE_SIZE];

	hi = port % 256;
	lo = (port - hi) / 256;

	replace_not_larger(addr, ".", ",");

	snprintf(buffer, sizeof(buffer), "PORT %s,%d,%d\r\n",
		addr, lo, hi);

	say(fd, buffer);
}

void destroy_active_portrange() {
	config_destroy_portrange(actvrangeserver);
	config_destroy_portrange(actvrangeclient);
	actvrangeserver = (struct portrangestruct*) 0;
	actvrangeclient = (struct portrangestruct*) 0;
}

