/* 
 * Copyright (C) 1999-2004 Joachim Wieland <joe@mcknight.de>
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place - Suite 330, Boston, MA 02111, USA.
 */

#include "ftp_main.h"
#include "ftp_login_cl.h"
#include <sys/mman.h>

#define ANON_USERS " anonymous ftp "

static int login_connect(struct proxyinfo*, struct messagebuffer*);
static int login_mayconnect(struct proxyinfo*, struct messagebuffer*);
static int login_init_connection(struct proxyinfo*, struct messagebuffer*);
static int login_connected_setup(struct proxyinfo*, struct messagebuffer*);

static int login_auth(struct proxyinfo*, struct messagebuffer*);
static int login_readwelcome(struct proxyinfo*, struct messagebuffer*);
static int login_sendauth_user(struct proxyinfo*, struct messagebuffer*);
static int login_sendauth_pass(struct proxyinfo*, struct messagebuffer*);
static int login_finish_login(struct proxyinfo*, struct messagebuffer*);
static int login_loggedin_setup(struct proxyinfo*, struct messagebuffer*);

static int login_connect_failed(struct proxyinfo*, struct messagebuffer*);
static int login_failed(struct proxyinfo*, struct messagebuffer*);


/* The function receives a stage that tells it until which step it should do
 * the login */

int login(struct proxyinfo* pi, struct messagebuffer* msgbuf, int stage) {
	int cs = pi->si->connection.clientsocket;
	struct clientinfo* ci = pi->ci;
	int ret = 0;

	/* we need a destination already for the most simple stage, so we need
	 * it for all stages */
	if (! ci->destination) {
		message_buffer_add(msgbuf, cs,
				500, "Error logging in\r\n");
		jlog(4, "No destination was set - Cannot proceed");
		if (config_compare_option("logintime", "connect")) {
			jlog(4, "This may be because of your logintime --> connect setting");
		}
		return CMD_ERROR;
	}

	/* if we should log into the remote server and send the username or
	 * even do more but we don't have a username already, that could be a
	 * little problem */
	if (! ci->user && stage >= LOGIN_ST_USER) {
		message_buffer_add(msgbuf, cs,
				500, "Error logging in\r\n");
		jlog(4, "No user was set - Cannot proceed");
		return CMD_ERROR;
	}

	if (stage >= LOGIN_ST_CONNECTED) {
		ret = login_connect(pi, msgbuf);
		if (ret) {
			int ret2;
			if ((ret2 = login_connect_failed(pi, msgbuf)) < 0) {
				return ret2;
			}
			return ret;
		}
	}
	if (stage >= LOGIN_ST_USER) {
		ret = login_sendauth_user(pi, msgbuf);
		if (ret) { return ret; }
	}
	if (stage >= LOGIN_ST_FULL) {
		ret = login_connect(pi, msgbuf);
		if (ret) {
			int ret2;
			if ((ret2 = login_connect_failed(pi, msgbuf)) < 0) {
				return ret2;
			}
			return ret;
		}
		ret = login_auth(pi, msgbuf);
		if (ret) {
			int ret2;
			if ((ret2 = login_failed(pi, msgbuf)) < 0) {
				return ret2;
			}
			return ret;
		} else {
			config_destroy_sectionconfig();
		}
		/* We could login successfully */
		/* set up everything */
		ret = login_loggedin_setup(pi, msgbuf);
		if (ret) { return ret; }
	}
	return CMD_HANDLED;
}


static
int login_connect(struct proxyinfo* pi, struct messagebuffer* msgbuf) {
	struct clientinfo* ci = pi->ci;
	int ret;

	if (ci->login.stage >= LOGIN_ST_CONNECTED) {
		return CMD_HANDLED;
	}

	/* open the connection and return the socket descriptor. Copy the
	 * descriptor to the pi->si structure as well */
	if ((ret = login_init_connection(pi, msgbuf)) < 0) {
		/* the error is logged and say()ed */
		return ret;
	}

	/* read the welcome message.
	 * If we should connect to the target server at connecttime, copy the
	 * welcomemessage to the pi->ci structure
	 * Send an initial SYST command and see how the server responds */
	if ((ret = login_connected_setup(pi, msgbuf)) < 0) {
		/* the error is logged and say()ed */
		return ret;
	}

	ci->login.stage = LOGIN_ST_CONNECTED;
	return CMD_HANDLED;
}


static
int login_auth(struct proxyinfo* pi, struct messagebuffer* msgbuf) {
	struct clientinfo* ci = pi->ci;
	int ret;

	if (ci->login.stage >= LOGIN_ST_LOGGEDIN) {
		return CMD_HANDLED;
	}

	if (ci->login.stage < LOGIN_ST_USER) {
		/* send the username, read the answer and save it to the pi->ci
		 * structure. */
		if ((ret = login_sendauth_user(pi, msgbuf)) < 0) {
			ci->login.stage = LOGIN_ST_CONNECTED;
			return ret;
		}
		/* XXX Set by login_sendauth_user as well */
		ci->login.stage = LOGIN_ST_USER;
	}

	/* Send the password - but don't read the answer */
	if ((ret = login_sendauth_pass(pi, msgbuf)) < 0) {
		ci->login.stage = LOGIN_ST_CONNECTED;
		return ret;
	}

	/* Read the answer of the server. Concatenate it with the welcome
	 * message if necessary and check the result code */
	if ((ret = login_finish_login(pi, msgbuf)) < 0) {
		ci->login.stage = LOGIN_ST_CONNECTED;
		return ret;
	}

	return CMD_HANDLED;
}


static
int login_mayconnect(struct proxyinfo* pi, struct messagebuffer* msgbuf) {
	int cs = pi->si->connection.clientsocket;
	struct clientinfo* ci = pi->ci;
	struct serverinfo* si = pi->si;

	int allowed = 0;
	unsigned long int target;
	unsigned long int host_ip;
	unsigned long int client_ip, server_ip;
	int tags;

	allowed = 0;

	/* we don't know yet if ci->destination is a hostname like
	 * "somehost.foo.com" or if it is an IP in a char* like
	 * "212.117.232.20"
	 *
	 * So call inet_addr(), it should make the decision  :-)
	 *
	 * */
	target = inet_addr(ci->destination);

	/* get the client_ip by asking information about the socket the
	 * client is connected */
	client_ip = get_uint_ip(GET_IP_CLIENT, pi);
	server_ip = (unsigned long int) UINT_MAX;

	if (target == (unsigned long int) UINT_MAX) {
		/* ci->destination may be a hostname, but it needn't.
		 * It may also be an invalid IP: "393.39.239.500" or
		 * anything else. Look it up.
		 * */
		host_ip = hostent_get_ip(&(si->hostcache), ci->destination);
		if (host_ip != (unsigned long int) UINT_MAX) {
			/* successful lookup */
			server_ip = host_ip;
		} else {
			/* it's not a valid IP and we could not look it up,
			 * so we could not get valid information about the
			 * destination host */
			jlog(7, "Nonsense destination (no IP and could not look up hostname): %s", ci->destination);
			jlog(8, "Please check your nameserver configuration. This may also happen if your chroot-environment does not contain the necessary files which the libc needs for a lookup");
			allowed = 0;
		}
	} else {
		/* okay, it's an IP */
		server_ip = inet_addr(ci->destination);
	}

	if (server_ip != (unsigned long) UINT_MAX) {
		/* we could get a valid IP, now we can evaluate if client
		 * is allowed to connect to the destination host */
		jlog(9, "Checking all tags");
		tags = TAG_ALL_NOT_FORWARDED;
		if (ci->forward.original.user) {
			/* if we are treating a forward, add this value, it
			 * won't come up in the option list otherwise */
			tags |= TAG_FORWARDED;
		}
		/* checking of the configuration works such:
		 *
		 * no forward:   do not evaluate <forwarded>
		 *               always check the normal values
		 *
		 * forward:      evaluate <forwarded>
		 *               if (!in_forwarded_tag) {
		 *     (1)             check original.* values
		 *               } else {
		 *     (2)             check normal values
		 *               }
		 */
		if (ci->forward.original.user) {
			/* Yes, there was a forward applied */
			config_shrink_config(client_ip,
				/* this are the original values */
				/* see above (1) */
				ci->forward.original.dest_ip,
				ci->forward.original.destination,
				ci->forward.original.destinationport,
				ci->forward.original.user,
				/* these are the values set by the forward */
				/* pass them as well, if we are in the
				 * forwarded_tag, config_match_section()
				 * will overwrite the previous with the
				 * following values */
				server_ip,
				ci->destination,
				ci->destinationport,
				ci->user,
				0,		/* set no specific time */
				si->connection.proxy_ip,
				si->connection.proxy_port,
				si->servertype,
				si->ready_to_serve,
				&si->hostcache,
				tags);
		} else {
			/* not a forward */
			config_shrink_config(client_ip,
				server_ip,
				ci->destination,
				ci->destinationport,
				ci->user,
				ci->forward.original.dest_ip,
				ci->forward.original.destination,
				ci->forward.original.destinationport,
				ci->forward.original.user,
				0,		/* set no specific time */
				si->connection.proxy_ip,
				si->connection.proxy_port,
				si->servertype,
				si->ready_to_serve,
				&si->hostcache,
				tags);
		}
		allowed = config_compare_option("access", "allow");
	}

	if (!allowed) {
		message_buffer_add(msgbuf, cs, 531,
			"You are not allowed to connect to that host.\r\n");
		jlog(8, "Not allowed to connect to %s", ci->destination);
		return CMD_ERROR;
	}
	/* if the client was allowed, save the clients original IP */
	if (!ci->forward.original.user) {
		ci->forward.original.dest_ip = server_ip;
	}
	return allowed; /* a positive int */
}


#define ERR_STR_P2 "Error connecting to %s port %d: %s\r\n"

static
int login_init_connection(struct proxyinfo* pi, struct messagebuffer* msgbuf) {
	int ss;
	int cs = pi->si->connection.clientsocket;
	struct clientinfo* ci = pi->ci;
	struct serverinfo* si = pi->si;
	struct ip_t srv_ip;
	const char* srv_name;
	int ret, err;

	if ((ret = login_mayconnect(pi, msgbuf)) < 0) {
		/* the error is logged and say()ed */
		return ret;
	}

	srv_ip = ip_parse_from_char(ci->destination);
	if (srv_ip.error) {
		/* it was not an ip - look up the name to get the ip */
		si->connection.server_ip = hostent_get_ip(&si->hostcache,
							ci->destination);
		/* if we have a hostname but can't look it up, we can never get
		 * here... normally... because we look it up to check access
		 * and stuff... XXX what about disabling lookups? what would
		 * happen? */
		ASSUME(si->connection.server_ip != ULONG_MAX);
		si->connection.server_name = strdup(ci->destination);
		enough_mem(si->connection.server_name);
	} else {
		struct in_addr addr;
		addr.s_addr = srv_ip.ip;
		srv_name = hostent_get_name(&si->hostcache,
						si->connection.server_ip);
		si->connection.server_ip = srv_ip.ip;
		if (!srv_name) {
			/* we can't look it up */
			si->connection.server_name = strdup(inet_ntoa(addr));
		} else {
			si->connection.server_name = strdup(srv_name);
		}
		enough_mem(si->connection.server_name);
	}

	/* connect */
	ss = openportname(ci->destination,     /* dest name */
			  ci->destinationport, /* dest port */
			                             /* source address */
			  config_get_addroption("controlserveraddress",
								INADDR_ANY),
			  (struct portrangestruct*) 0, /* source port */
			  pi);
	if (ss < 0) {
		err = errno;
		message_buffer_addf(msgbuf, cs,
				500, ERR_STR_P2,
				ci->destination,
				ci->destinationport,
				strerror(err));
		/* openportname() has already done a jlog() for this */
		return CMD_ERROR;
	}

	/* connected */
	pi->si->connection.serversocket = ss;

	return CMD_HANDLED;
}
#undef ERRSTR


static
int login_connected_setup(struct proxyinfo* pi, struct messagebuffer* msgbuf) {
	int ss = pi->si->connection.serversocket;
	int cs = pi->si->connection.clientsocket;

	int ret;
	struct ftp_message msg;

	if ((ret = login_readwelcome(pi, msgbuf)) < 0) {
		return ret;
	}

	if (config_get_bool("initialsyst", 1) == 1) {
		/* XXX */
		say(ss, "SYST\r\n");
		msg = ftp_readall(ss);
		if (msg.error) {
			ftp_message_say_error(msg, msgbuf, cs,
							__FILE__, __LINE__);
			return CMD_ABORT;
		}
		if (!strlen(msg.msg)) {
			jlog(2, "The server did not respond to the initial SYST command");
			message_buffer_add(msgbuf, cs, 500,
				"The server did not respond correctly\r\n");
			ftp_message_free(&msg);;
			return CMD_ERROR;
		}
		/* XXX check for a return code is missing */
		ftp_message_free(&msg);
	} else {
		jlog(9, "Suppressing the initial SYST command");
	}

	return CMD_HANDLED;
}


static
int login_readwelcome(struct proxyinfo *pi, struct messagebuffer* msgbuf) {
	int ss = pi->si->connection.serversocket;
	int cs = pi->si->connection.clientsocket;
	struct clientinfo* ci = pi->ci;
	/* Read the welcome line */
	struct ftp_message msg;
	char* msg_no_codes;

	msg = ftp_readall(ss);
	if (msg.error) {
		ftp_message_say_error(msg, msgbuf, cs, __FILE__, __LINE__);
		return CMD_ABORT;
	}

	jlog(9, "Connected to %s, got \"%s\" as welcome message",
		ci->destination, msg.msg);

	msg_no_codes = strip_respcodes(msg.msg);

	if (msg.code != 220) {
		jlog(2, "Not a valid FTP server response (%s)", msg.msg);
		message_buffer_add_nomalloc(msgbuf, cs,
				500, msg_no_codes);
		ftp_message_free(&msg);
		return CMD_ERROR;
	} else {
		ci->login.msg_welcome = msg_no_codes;
		ftp_message_free(&msg);
	}
	return CMD_HANDLED;
}


static
int login_sendauth_user(struct proxyinfo* pi, struct messagebuffer* msgbuf) {
	int ss = pi->si->connection.serversocket;
	int cs = pi->si->connection.clientsocket;
	struct clientinfo* ci = pi->ci;
	struct ftp_message msg;

	int ret;
	char* sendbuf, *msg_no_codes;

	if (ci->login.stage >= LOGIN_ST_USER) {
		return CMD_HANDLED;
	}

	sendbuf = char_enclose("USER ", ci->user, "\r\n");
	ret = say(ss, sendbuf);
	free(sendbuf);
	if (ret < 0) {
		jlog(2, "Error writing the user name to the server: %s",
			strerror(errno));
		return CMD_ABORT;
	}

	msg = ftp_readall(ss);

	if (msg.error) {
		ftp_message_say_error(msg, msgbuf, cs, __FILE__, __LINE__);
		return CMD_ABORT;
	}

	msg_no_codes = strip_respcodes(msg.msg);

	if (msg.code != 331 && msg.code != 230) {
		/* prepend the response from the server with an error code
		 * and return it to the client */
		int code = 500;
		if (msg.code == 530) { code = 530; }
		message_buffer_add_nomalloc(msgbuf, cs,
						code, msg_no_codes);

		jlog(7, "Got \"%s\" after sending the username.", msg.msg);
		ftp_message_free(&msg);
		return CMD_ERROR;
	}
	if (msg.code == 230) {
		ci->login.stage = LOGIN_ST_LOGGEDIN;
	} else {
		/* msg.code == 331 */
		ci->login.stage = LOGIN_ST_USER;
	}
	ci->login.msg_after_user = msg_no_codes;
	ftp_message_free(&msg);
	return CMD_HANDLED;
}


static
int login_sendauth_pass(struct proxyinfo* pi, struct messagebuffer* msgbuf) {
	int ss = pi->si->connection.serversocket;
	struct clientinfo* ci = pi->ci;

	int ret;
	char* sendbuf;
	char* userdup;

	sendbuf = char_enclose("PASS ", ci->pass, "\r\n");
	ret = say(ss, sendbuf);

	userdup = char_enclose(" ", ci->user, " ");
	if (strstr(ANON_USERS, userdup)) {
		ci->anon_user_pass = ci->pass;
		ci->pass = (char*) 0;
	} else {
		/* XXX caching */
		overwrite_string(sendbuf);
		ci->anon_user_pass = strdup("UNKNOWN");
	}
	free(sendbuf);
	free(userdup);

	if (ret < 0) {
		jlog(2, "Error writing the password to the server: %s",
			strerror(errno));
		return CMD_ABORT;
	}
	
	return CMD_HANDLED;
}


static
int login_loggedin_setup(struct proxyinfo* pi, struct messagebuffer* msgbuf) {
	int ss = pi->si->connection.serversocket;
	struct clientinfo* ci = pi->ci;
	struct serverinfo* si = pi->si;

	/* we seem to have a successful login */
	jlog(7, "Logged in to %s as %s!", ci->destination, ci->user);

	/* we don't need the configuration sections and the backup anymore */
	config_delete_master();
	config_delete_backup();

	/* get the throughput rate */
	si->throughput = config_get_foption("throughput", -1.0);

	si->connection.addr_to_server = socketinfo_get_local_ip(ss);

	/* XXX data type address_t or ip_t */
	ci->data_connection.data_addr_to_client
				= config_get_addroption("dataclientaddress",
						si->connection.addr_to_client);
	jlog(8, "Got %s as data client address",
	       inet_ntoa(*((struct in_addr*) &ci->data_connection.data_addr_to_client)));

	ci->data_connection.data_addr_to_server =
				config_get_addroption("dataserveraddress",
						si->connection.addr_to_server);
	jlog(8, "Got %s as data server address",
	       inet_ntoa(*((struct in_addr*) &ci->data_connection.data_addr_to_server)));

	ci->login.stage = LOGIN_ST_FULL;
	event_fire(EVENT_FTP_LOGGEDIN, pi);
	return CMD_HANDLED;
}


static
int login_finish_login(struct proxyinfo* pi, struct messagebuffer* msgbuf) {
	int ss = pi->si->connection.serversocket;
	int cs = pi->si->connection.clientsocket;
	struct clientinfo* ci = pi->ci;
	struct ftp_message msg;
	char* msg_no_codes;

	if (stage_action("loggedin", pi) < 0) {
		message_buffer_add(msgbuf, cs,
				421, "Error setting up (see logfile)\r\n");
		return CMD_ABORT;
	}

	msg = ftp_readall(ss);
	if (msg.error) {
		ftp_message_say_error(msg, msgbuf, cs, __FILE__, __LINE__);
		return CMD_ABORT;
	}

	msg_no_codes = strip_respcodes(msg.msg);
	if (msg.code != 230) {
		message_buffer_add_nomalloc(msgbuf, cs,
				530, msg_no_codes);
		jlog(8, "Got \"%s\" after sending the password", msg.msg);
		ftp_message_free(&msg);
		return CMD_ERROR;
	}
	ci->login.msg_after_pass = msg_no_codes;
	ftp_message_free(&msg);
	return CMD_HANDLED;
}


static
int login_connect_failed(struct proxyinfo* pi, struct messagebuffer* msgbuf) {
	struct serverinfo* si = pi->si;

	free_and_zero_char(&si->connection.server_name);
	si->connection.server_ip = ULONG_MAX;
	
	return CMD_HANDLED;
}

static
int login_failed(struct proxyinfo* pi, struct messagebuffer* msgbuf) {
	struct clientinfo* ci = pi->ci;
	struct serverinfo* si = pi->si;

	static int failed_logins;

	/* if there was a forward, restore the ci fields from the
	 * original values */

	if (ci->forward.original.user) {
		free(ci->user);
		ci->user = ci->forward.original.user;
		ci->forward.original.user = (char*) 0;
	}
	if (ci->forward.original.destination) {
		free(ci->destination);
		ci->destination = ci->forward.original.destination;
		ci->forward.original.destination = (char*) 0;
		ci->destinationport = ci->forward.original.destinationport;
	}

	failed_logins ++;

	/* XXX missing in documentation */
	if (failed_logins >= config_get_ioption("failedlogins", 3)) {
		return CMD_ABORT;
	}

	/* switch configuration - remove the shrinked one,
	 * set the backup into place. If the login failed,
	 * the backup is used */

	if (config_activate_backup() < 0) {
		jlog(6, "Backup could not be activated - can't continue with login procedure");
		return CMD_ABORT;
	}

	/* shrink the configuration again - this will update the option list */

	config_shrink_config(get_uint_ip(GET_IP_CLIENT, pi),
			(unsigned long int) UINT_MAX,
			(char*) 0,   /* destination */
			0,           /* destinationport */
			(char*) 0,   /* user */
			(unsigned long int) UINT_MAX,
			(char*) 0,   /* original.destination */
			0,           /* original.destinationport */
			(char*) 0,   /* original.user */
			0,           /* set no specific time */
			si->connection.proxy_ip,
			si->connection.proxy_port,
			si->servertype,
			si->ready_to_serve,
			&si->hostcache,
			TAG_CONNECTED);

	return CMD_HANDLED;
}


