/*
 * Copyright (C) 1999-2004 Joachim Wieland <joe@mcknight.de>
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place - Suite 330, Boston, MA 02111, USA.
 */

#include "ftp_main.h"
#include "ftp_cmds.h"
#include "ftp_std.h"

int cmds_after_client_connect(struct proxyinfo*, struct messagebuffer*);

/* static functions here */
static int cmds_after_client_user(struct proxyinfo*, struct messagebuffer*);
static int cmds_after_client_pass(struct proxyinfo*, struct messagebuffer*);
static int cmds_after_auth_user(struct proxyinfo*, struct messagebuffer*);
static int cmds_after_auth_pass(struct proxyinfo*, struct messagebuffer*);

static int try_forward(struct proxyinfo*, struct messagebuffer*, int);
static int try_forward_pass_rewrite(struct proxyinfo*, struct messagebuffer*);
static int try_transparent_forward(struct proxyinfo*,
					struct messagebuffer*, int);
static int try_transparent(struct proxyinfo*, struct messagebuffer*);

static int fw_validate(const char* fwuser, const char* fwpass,
							struct proxyinfo* pi);
static void fw_say_welcome(struct proxyinfo* pi, struct messagebuffer* msgbuf);

/* XXX exclude somehow special combinations i.e. login type 9 and a logintime =
 * connect forward... */

/* XXX test all functions for invalid input and let it check the parameters */

/* XXX exclude defaultforward for all loginstyles except for 1 */

#define L_NO		-1
#define L_UNSPEC	0
#define L_CONNECT	1
#define L_USER		2
#define L_PASS		3
static
int logintimes_valid[][3] = {
/*        CONNECT  USER  PASS  */
/* 0 */ {    1 ,     1 ,   1 }, /* NO rewriting */
/* 1 */ {    0 ,     1 ,   1 }, /* user@host:port */
/* 2 */ {    0 ,     1 ,   1 }, /* USER with login */
/* 3 */ {    0 ,     1 ,   1 }, /* SITE with login */
/* 4 */ {    0 ,     1 ,   1 }, /* SITE without login */
/* 5 */ {    0 ,     1 ,   1 }, /* OPEN with login */
/* 6 */ {    0 ,     1 ,   1 }, /* OPEN without login */
/* 7 */ {    0 ,     0 ,   1 }, /* user@fwuser@real.host.name, pass@fwpass */
/* 8 */ {    0 ,     1 ,   1 }, /* fwuser@real.host.name,fwpass,user,pass */
/* XXX actually PASS is too early as well */
/* 9 */ {    0 ,     0 ,   1 }}; /* User@host FireID, ACCT at last */

static
int is_logintime_too_early(void) {
	int loginstyle = config_get_ioption("loginstyle", -1);
	const char* logintime = config_get_option("logintime");
	int logintimes_validity[3];
	int invalid = 0;

	ASSUME(loginstyle >= 0);
	ASSUME(loginstyle <= 9);
	ASSUME(logintime != NULL);

	logintimes_validity[0] = logintimes_valid[loginstyle][0];
	logintimes_validity[1] = logintimes_valid[loginstyle][1];
	logintimes_validity[2] = logintimes_valid[loginstyle][2];

	if (strcmp(logintime, "connect") == 0) {
		if (!logintimes_validity[0]) {
			invalid = 1;
		}
	} else if (strcmp(logintime, "user") == 0) {
		if (!logintimes_validity[1]) {
			invalid = 1;
		}
	} else if (strcmp(logintime, "pass") == 0) {
		if (!logintimes_validity[2]) {
			invalid = 1;
		}
	} else {
		jlog(3, "logintime unknown: %s", logintime);
		return CMD_ABORT;
	}
	if (invalid) {
		jlog(4, "logintime %s too early for loginstyle %d",
						logintime, loginstyle);
		return CMD_ERROR;
	}
	return CMD_HANDLED;
}

int cmds_after_client_connect(struct proxyinfo* pi,
				struct messagebuffer* msgbuf) {
	struct clientinfo* ci = pi->ci;
	int cs = pi->si->connection.clientsocket;
	int ret;

	if ((ret = is_logintime_too_early())) {
		message_buffer_add(msgbuf, cs,
			500, "Internal configuration error\r\n");
		return ret;
	}
	if (config_compare_option("logintime", "connect")) {
		ret = try_forward(pi, msgbuf, LOGIN_ST_CONNECTED);
		if (ret) {
			return CMD_ERROR;
		}

		ret = try_transparent_forward(pi, msgbuf, LOGIN_ST_CONNECTED);
		if (ret) {
			return CMD_ERROR;
		}

		ret = try_transparent(pi, msgbuf);
		if (ret) {
			return CMD_ERROR;
		}

		if (ci->destination) {
			/* do the server login until the LOGIN_ST_USER stage */
			ret = login(pi, msgbuf, LOGIN_ST_CONNECTED);
			if (ret) { return CMD_ERROR; }
			/* welcomemessage */
			message_buffer_add(msgbuf, cs,
					220, ci->login.msg_welcome);
			free_and_zero_char(&ci->login.msg_welcome);
			return CMD_HANDLED;
		}
	}
	fw_say_welcome(pi, msgbuf);
	return CMD_HANDLED;
}

static
int cmds_after_client_user(struct proxyinfo* pi,
				struct messagebuffer* msgbuf) {
	int cs = pi->si->connection.clientsocket;
	struct clientinfo* ci = pi->ci;
	int ret;

	if ((ret = is_logintime_too_early())) {
		/* 550 because this is a failed USER command */
		message_buffer_add(msgbuf, cs,
			550, "Internal configuration error\r\n");
		return ret;
	}

	/* XXX 1 see if this is unnecessary - we have one in client_connect
	 * already. HOwever it might be possible to change a configuration
	 * setting
	 *
	 * <general>
	 *   logintime pass
	 *   ...
	 * </general>
	 *
	 * <user *>
	 *  logintime connect
	 * </user> */

	ret = try_forward(pi, msgbuf, LOGIN_ST_USER);
	if (ret) {
		return CMD_ERROR;
	}

	if (config_compare_option("logintime", "pass")) {
		/* we don't have to do special things here, just display a
		 * short message... */
		const char* user;

		/* display the user that logged in originally. The client
		 * doesn't need to see that he got mapped to another user */
		if (ci->forward.original.user) {
			user = ci->forward.original.user;
		} else {
			user = ci->user;
		}
		message_buffer_addf(msgbuf, cs,
			331, "Password required for %s.\r\n", user);

		return CMD_HANDLED;
	}

	/* either logintime = "connect" or logintime = "user" is set */
	ret = try_forward(pi, msgbuf, LOGIN_ST_CONNECTED);
	if (ret) {
		return CMD_ERROR;
	}

	ret = try_transparent_forward(pi, msgbuf, LOGIN_ST_CONNECTED);
	if (ret) {
		return CMD_ERROR;
	}

	ret = try_transparent(pi, msgbuf);
	if (ret) {
		return CMD_ERROR;
	}

	/* connect if we should connect after having received the USER command.
	 * Do not connect, if we are already connected or should connect later
	 * on */

	/* do the server login until the LOGIN_ST_USER stage */
	ret = login(pi, msgbuf, LOGIN_ST_USER);

	if (ret) { return CMD_ERROR; }

	if (ci->login.msg_welcome) {
		/* we haven't sent the welcome message yet */
		message_buffer_add(msgbuf, cs,
				331, ci->login.msg_welcome);
		free_and_zero_char(&ci->login.msg_welcome);
	}

	if (ci->login.stage == LOGIN_ST_LOGGEDIN) {
		message_buffer_add(msgbuf, cs,
				230, ci->login.msg_after_user);
		free_and_zero_char(&ci->login.msg_after_user);
		/* already logged in - without having to supply a password */
		if (login(pi, msgbuf, LOGIN_ST_FULL) < 0) {
			return CMD_ERROR;
		}
		/* this is the only case where we return CMD_DONE so
		 * that the login handler doesn't wait for the
		 * command that would follow but knows that his job
		 * is finished */
		return CMD_DONE;
	}

	ASSUME(ci->login.msg_after_user != NULL);
	message_buffer_add(msgbuf, cs,
			331, ci->login.msg_after_user);
	free_and_zero_char(&ci->login.msg_after_user);

	return CMD_HANDLED;
}

static
int cmds_after_client_pass(struct proxyinfo* pi,
				struct messagebuffer* msgbuf) {
	int ret;
	struct clientinfo* ci = pi->ci;
	int cs = pi->si->connection.clientsocket;

	/* XXX see XXX 1 */
	if ((ret = is_logintime_too_early())) {
		/* 550 because this is a failed USER command */
		message_buffer_add(msgbuf, cs,
			550, "Internal configuration error\r\n");
		return ret;
	}

	if (try_forward_pass_rewrite(pi, msgbuf) != CMD_HANDLED) {
		return CMD_ERROR;
	}

	ret = try_transparent_forward(pi, msgbuf, LOGIN_ST_CONNECTED);
	if (ret) {
		return CMD_ERROR;
	}

	ret = try_transparent(pi, msgbuf);
	if (ret) {
		return CMD_ERROR;
	}

	/* connect to the remote server and log in. We now have all the
	 * necessary data */
	ret = login(pi, msgbuf, LOGIN_ST_FULL);
	if (ret == CMD_ABORT) {
		return CMD_ABORT;
	}
	if (ret < 0) {
		/* login failed */
		return CMD_ERROR;
	}
	if (ci->login.msg_welcome) {
		/* we haven't sent the welcome message yet */
		message_buffer_add(msgbuf, cs,
				331, ci->login.msg_welcome);
		free_and_zero_char(&ci->login.msg_welcome);
	}
	if (ci->login.msg_after_user) {
		int code = ci->login.stage >= LOGIN_ST_LOGGEDIN ? 230 : 331;
		message_buffer_add(msgbuf, cs,
				code, ci->login.msg_after_user);
		free_and_zero_char(&ci->login.msg_after_user);
	}
	if (ci->login.msg_after_pass) {
		/* this is NULL if we just got a 230 after the user name */
		message_buffer_add(msgbuf, cs,
				230, ci->login.msg_after_pass);
		free_and_zero_char(&ci->login.msg_after_pass);
	}
	return CMD_HANDLED;
}


static
int cmds_after_auth_user(struct proxyinfo* pi,
				struct messagebuffer* msgbuf) {
	struct clientinfo* ci = pi->ci;

	jlog(8, "User supplied auth username '%s'", ci->login.fw_auth.user);
	/* at the moment this function returns always success
	 * (this comment included in other places here as well in this file) */
	return CMD_HANDLED;
}

static
int cmds_after_auth_pass(struct proxyinfo* pi,
				struct messagebuffer* msgbuf) {
	int cs = pi->si->connection.clientsocket;
	struct clientinfo* ci = pi->ci;

	if (ci->login.fw_auth.user && ci->login.fw_auth.pass &&
	    fw_validate(ci->login.fw_auth.user,
		    		ci->login.fw_auth.pass, pi) == AUTH_OK) {
		message_buffer_add(msgbuf, cs,
				230, "Login to proxy successful\r\n");
	} else {
		sleep(3);
		message_buffer_add(msgbuf, cs,
				530, "Login failed.\r\n");
		return CMD_ERROR;
	}
	return CMD_HANDLED;
}

static
int parse_login_string(const char* login, struct ftp_login_t* login_values) {
	char* portstr, *modestr;
	int offset = 0;

	login_values->user = (char*) 0;
	login_values->destination = (char*) 0;
	login_values->destinationport = 0;
	login_values->mode = UNSPEC;

	/* joe@host,21,p */
	/* joe */
	/* joe@host */
	/* joe@host:21 */
	/* joe,host,21,p */

	if (!login) {
		return -1;
	}
	if (!login[0]) {
		return -1;
	}
	if (login[0] == '@') {
		return -1;
	}
	if (login[strlen(login)-1] == '@') {
		return -1;
	}
	if (!strchr(login, '@') && !strchr(login, ',')) {
		/* Only the user name: joe */
		login_values->user = quotstrtok(login, " \t", &offset);
		return 0;
	}
	
	/* always include the "own" delimiter in order to strip this symbol */
	if (!strchr(login, '@')) {
		/* we parse only destination, destinationport and mode flags */
	} else {
		login_values->user = quotstrtok(login, "@ \t", &offset);
	}
	login_values->destination = quotstrtok(login, "@:, \t", &offset);
	portstr = quotstrtok(login, ":, \t", &offset);
	modestr = quotstrtok(login, ", \t", &offset);

	if (portstr) {
		long int pno;
		pno = strtol(portstr, NULL, 10);
		if ((errno == ERANGE && (pno == LONG_MIN || pno == LONG_MAX))
				|| pno == 0) {
			/* it was not a number */
			login_values->destinationport = 0;
			/* maybe it is a mode */
			/* XXX it is an error if mode is set here */
			free(modestr);
			modestr = portstr;
			portstr = (char*) 0;
		} else {
			/* it was a number */
			login_values->destinationport = pno;
			free(portstr);
		}
	}

	if (modestr) {
		if (modestr && (strchr("ap", *modestr))) {
			jlog(9, "mode specified: %s", modestr);
			switch (*modestr) {
			case 'a':
				login_values->mode = ACTIVE;
				jlog(9, "p-s: active");
				break;
			case 'p':
				login_values->mode = PASSIVE;
				jlog(9, "p-s: passive");
				break;
			}
		}
	}
	return 0;
}


static
int parse_login_string_without_user(const char* login,
					struct ftp_login_t* login_values) {
	char* new_login;
	int i;
	if (!login) {
		return -1;
	}
	if (strchr(login, '@')) {
		return -1;
	}
	new_login = char_prepend("dummy@", login);
	i = parse_login_string(new_login, login_values);
	free(new_login);
	if (strcmp(login_values->user, "dummy") == 0) {
		free_and_zero_char(&login_values->user);
	}
	return i;
}

static
void get_login_values(struct proxyinfo* pi, struct ftp_login_t* lv) {
	struct clientinfo* ci = pi->ci;

	/* if we have logintime connect, we might not yet have a user
	 * name */
	lv->user = strnulldup(ci->user);
	lv->destination = strdup(ci->destination);
	enough_mem(lv->destination);
	lv->destinationport = ci->destinationport;
	lv->mode = ci->data_connection.transfermode.servermode;
}

static
void set_login_values(struct proxyinfo* pi, struct ftp_login_t* lv) {
	struct clientinfo* ci = pi->ci;

	if (ci->user) { free(ci->user); }
	ci->user = lv->user;
	if (ci->destination) { free(ci->destination); }
	ci->destination = lv->destination;
	ci->destinationport = lv->destinationport;
	ci->data_connection.transfermode.servermode = lv->mode;
}

static
void set_login_values_if_undefined(struct proxyinfo* pi,
					struct ftp_login_t* lv) {
	struct clientinfo* ci = pi->ci;

	if (!ci->user) { ci->user = lv->user; }
		else { free(lv->user); }
	if (!ci->destination) { ci->destination = lv->destination; }
		else { free(lv->destination); }
	if (!ci->destinationport) { ci->destinationport = lv->destinationport; }
	if (ci->data_connection.transfermode.servermode == UNSPEC) {
		ci->data_connection.transfermode.servermode = lv->mode;
	}
}

static
void apply_default_login_values(struct ftp_login_t* login_values) {
	if (login_values->destinationport == 0) {
		login_values->destinationport
			= config_get_ioption("serverport", DEFAULTSERVERPORT);
	}
	if (login_values->mode == UNSPEC) {
		if (config_compare_option("defaultmode", "active")) {
			login_values->mode = ACTIVE;
		} else if (config_compare_option("defaultmode", "passive")) {
			login_values->mode = PASSIVE;
		} else if (config_compare_option("defaultmode", "asclient")) {
			login_values->mode = ASCLIENT;
		} else {
			/* default fallback */
			login_values->mode = ASCLIENT;
		}
	}
}

static
void apply_missing_login_values(struct proxyinfo* pi) {
	struct ftp_login_t login_values;

	get_login_values(pi, &login_values);
	apply_default_login_values(&login_values);
	set_login_values(pi, &login_values);
}


static
int try_forward(struct proxyinfo* pi, struct messagebuffer* msgbuf, int stage) {
	int cs = pi->si->connection.clientsocket;
	struct clientinfo* ci = pi->ci;
	struct serverinfo* si = pi->si;

	struct ftp_login_t login_values;
	const char* forward;
	int loginstyle;
	struct slist_t* forward_list, *forward_list_cur;
	int ret;
	/* if we have a forward, we can either change user/destination/...
	 * directly and indicate that by setting forward_applied to 1.
	 * However for most situations we can leave it at 0 and apply it at the
	 * end of this function. We change forward.login to reflect the action
	 * of the forward */
	int forward_applied = 0;
	int config_state = TAG_GLOBAL | TAG_FROM | TAG_PORT
			 | TAG_TIME | TAG_SERVERTYPE
			 | TAG_PROXYIP | TAG_PROXYPORT;

	/* this is already a forward */
	if ( ci->forward.active ) {
		return CMD_HANDLED;
	}

	if ( ci->destination ) {
		config_state |= TAG_TO;
	}

	if ( ci->user ) {
		config_state |= TAG_USER;
	}

	config_shrink_config(get_uint_ip(GET_IP_CLIENT, pi),
			get_uint_ip(GET_IP_SERVER, pi),
			ci->destination,
			ci->destinationport,
			ci->user,
			-1,          /* original.dest_ip */
			(char*) 0,   /* original.destination */
			0,           /* original.destinationport */
			(char*) 0,   /* original.user */
			0,           /* set no specific time */
			si->connection.proxy_ip,
			si->connection.proxy_port,
			si->servertype,
			si->ready_to_serve,
			&si->hostcache,
			config_state);

	if (config_compare_option("access", "allow") == 0) {
		message_buffer_add(msgbuf, cs, 531,
			"You are not allowed to connect to that host.\r\n");
		jlog(5, "%s was not allowed to connect.",
				conv_ip2char(si->connection.client_ip));
		return CMD_ERROR;
	}

	/* see if there is an option for the forward */
	forward = config_get_option("forward");
	loginstyle = config_get_ioption("loginstyle", -1);

	if (!forward) {
		/* no forward */
		return CMD_HANDLED;
	}

	/*
	 * <user ftp>
	 *    forward johnfred@fooserver.com
	 *    forward johnfred@fooserver.com,3949,p
	 *    forward johnfred@fooserver.com     *     johnspass
	 *    forward johnfred@fooserver.com   JiKe94  johnspass
	 * </user>
	 *    forward *@fooserver.com:2378
	 *    forward %@fooserver.com:2378
	 */

	/* split the list into tokens */
	if ( ! (forward_list = config_split_line( forward, WHITESPACES )) ) {
		return CMD_HANDLED;
	}

	/* check the forward we got for compatibility with the environment, we
	 * do it after the config_split_line in order to have it trimmed and
	 * split up... */

	/* we do allow forwards like *@somewhere and %@somewhere only if we are
	 * in loginstyle 1 */

	if (loginstyle != 1) {
		if (forward_list->value &&
		   strlen(forward_list->value) >= 2 &&
		   (forward_list->value[0] == '%'
		       || forward_list->value[0] == '*') &&
		   forward_list->value[1] == '@') {

			slist_destroy_free(forward_list);
			message_buffer_add(msgbuf, cs,
				500, "Internal configuration error\r\n");
			jlog(4, "loginstyle %d does not support forwards like "
					"'*@...' or '%@...'", loginstyle);
			return CMD_ERROR;
		}
	}

	/* if there is a defaultforward but we already have a destination
	 * set, we can quit here as well */
	if (ci->destination &&
	    forward_list->value &&
	    strlen(forward_list->value) >= 2 &&
	    forward_list->value[0] == '%' &&
	    forward_list->value[1] == '@') {
		/* no forward is being used */
		slist_destroy_free(forward_list);
		jlog(8, "ignoring defaultforward %s "
			"- we have a destination already", forward);
		return CMD_HANDLED;
	}

	/* now we have a forward - and we use it */
	ci->forward.active = 1;
	forward_list_cur = forward_list;

	/* read and save the values */
	ci->forward.login = strdup(forward_list_cur->value);
	enough_mem(ci->forward.login);

	forward_list_cur = forward_list_cur->next;

	if (forward_list_cur && forward_list_cur->value) {
		ci->forward.accept_pw = strdup(forward_list_cur->value);
		enough_mem(ci->forward.accept_pw);
	} else {
		ci->forward.accept_pw = (char*) 0;
	}

	if (forward_list_cur) {
		forward_list_cur = forward_list_cur->next;
	}
	if (forward_list_cur && forward_list_cur->value) {
		ci->forward.send_pw = strdup(forward_list_cur->value);
		enough_mem(ci->forward.send_pw);
	} else {
		ci->forward.send_pw = (char*) 0;
	}
	/* destroy the list - we have saved the values into
	 * ci->forward.<field>    */
	slist_destroy_free(forward_list);

	/* delete those configuration values that are evaluated by the
	 * parsing routine  -  they are not necessary anymore */
	config_option_list_delete("transparent-forward");
	config_option_list_delete("forward");

	/* back up - ci->user and ci->destination don't have to be set. Hence
	 * original.* will stay NULL as well and will be set later on */
	ci->forward.original.user = strnulldup(ci->user);
	ci->forward.original.destination = strnulldup(ci->destination);
	/* XXX check if the above applies here as well */
	ci->forward.original.destinationport = ci->destinationport;
	ci->forward.original.dest_ip = get_uint_ip(GET_IP_SERVER, pi);

	/* this part will be executed for a cmds_after_client_connect()
	 * having a default forward like this:
	 *
	 * logintime connect
	 * forward somewhere.else.com
	 */
	if (!strchr(ci->forward.login, '@') && !ci->destination) {
		struct ftp_login_t login_values;
		int ret = parse_login_string_without_user(ci->forward.login,
							&login_values);
		if (ret) {
			/* XXX destruct routine allocs */
			return CMD_ERROR;
		}
		set_login_values_if_undefined(pi, &login_values);
		apply_missing_login_values(pi);
		forward_applied = 1;
	}
	
	if (stage < LOGIN_ST_USER) {
		/* XXX destruct routine allocs */
		return CMD_HANDLED;
	}

	/* here ci->user is set and so is ci->forward.original.user */

	/* if there was only a new destination but no new username given,
	 * prepend the old one */
	if (!strchr(ci->forward.login, '@')) {
		char* new;
		char* esc_user = escape_username(ci->forward.original.user);
		new = char_enclose(esc_user,
					"@",
					ci->forward.login);
		free(ci->forward.login);
		free(esc_user);
		ci->forward.login = new;

		/* reset the destination - we'll have a new one */
		free_and_zero_char(&ci->destination);
		/* forward_applied stays at 0 */
	}

	if (ci->forward.login[0] == '*' &&
	    ci->forward.login[1] == '@') {
		/* passauth */
		ci->forward.passauth = 1;

		/* call the parsing routine for the forward setting and let it
		 * set the values */
		ret = parse_login_string(ci->forward.login, &login_values);
		if (ret) {
			/* XXX destruct routine allocs */
			return CMD_ERROR;
		}

		/* however it has set * as the user name - replace it by the
		 * verbatim given user (passauth) */
		free(login_values.user);
		login_values.user = strdup(ci->user_login);
		enough_mem(login_values.user);

		apply_default_login_values(&login_values);
		set_login_values(pi, &login_values);
		forward_applied = 1;
	}

	/* If there is still no destination, see if we have a defaultforward
	 * setting */
	if (ci->destination == (char*) 0 &&
	    ci->forward.login[0] == '%' &&
	    ci->forward.login[1] == '@') {
		/* defaultforward */
		/* XXX do we need to free ci->forward.login ? */
		ci->forward.login = char_append(ci->user,
						&(ci->forward.login[1]));

		jlog(8, "No destination was set. Using %s because of defaultforward setting", ci->forward.login);
		/* forward_applied stays at 0 */
	}

	/* we haven't applied a forward yet, but he have one defined. apply it
	 * now */
	if (!forward_applied) {
		/* call the parsing routine and let it set the values */
		ret = parse_login_string(ci->forward.login, &login_values);
		if (ret) {
			/* XXX destruct routine allocs */
			return CMD_ERROR;
		}
		apply_default_login_values(&login_values);
		set_login_values(pi, &login_values);
	}
	/* we don't have to free the values from login_values because
	 * they got copied to ci */

	/* shrink the configuration again - with TAG_FORWARDED this time */
	config_shrink_config(get_uint_ip(GET_IP_CLIENT, pi),
			/* these are the original values */
			/* see above (1) */
			ci->forward.original.dest_ip,
			ci->forward.original.destination,
			ci->forward.original.destinationport,
			ci->forward.original.user,
			/* these are the values set by the forward */
			/* pass them as well, if we are in the
			 * forwarded_tag, config_match_section()
			 * will overwrite the previous with the
			 * following values */
			get_uint_ip(GET_IP_SERVER, pi),
			ci->destination,
			ci->destinationport,
			ci->user,

			0,		/* set no specific time */

			si->connection.proxy_ip,
			si->connection.proxy_port,

			si->servertype,
			si->ready_to_serve,
			&si->hostcache,
			TAG_ALL);

	/* XXX destruct routine allocs */
	return CMD_HANDLED;
}


static
int try_forward_pass_rewrite(struct proxyinfo* pi,
					struct messagebuffer* msgbuf) {
	int cs = pi->si->connection.clientsocket;
	struct clientinfo* ci = pi->ci;
	/* check if the password matches */
	if (ci->forward.accept_pw) {
		/* allow any pw but use destpw on dest */
		if (strcmp(ci->forward.accept_pw, "*") == 0) {
			jlog(7, "Allowing any password because forward option "
					"has set password to '*'");
		}
		if (strcmp(ci->forward.accept_pw, "*") == 0
			/* check for the exact password */
			|| cryptcmp(ci->forward.accept_pw, ci->pass) == 0) {

			jlog(8, "The user sent the expected password that "
				"the proxy got from the forward option");
			free(ci->pass);
			ci->pass = strdup(ci->forward.send_pw);
			enough_mem(ci->pass);
		} else if (cryptcmp(ci->forward.accept_pw,
						ci->pass) != 0){
			jlog(7, "Expecting the password set in the "
				"forward option but the user failed to "
				"send the correct password");
			message_buffer_add(msgbuf, cs,
					530, "Login incorrect\r\n");
			sleep(3);
			return CMD_ERROR;
		} else {
			/* we should never get here */
			ASSUME(0);
		}
	}
	return CMD_HANDLED;
}


static
int try_transparent(struct proxyinfo* pi, struct messagebuffer* msgbuf) {
	int cs = pi->si->connection.clientsocket;
	struct clientinfo* ci = pi->ci;
	struct serverinfo* si = pi->si;
	struct address_t t_destination;

	if (si->connection.transparent == TRANSPARENT_NO) {
		/* not in transparent mode */
		jlog(9, "try_transparent: not in transparent mode");
		return CMD_HANDLED;
	}

	if (config_get_option("transparent-forward")) {
		/* transparent forward active */
		jlog(9, "try_transparent: transparent forward active");
		return CMD_HANDLED;
	}

	if (ci->destination) {
		jlog(9, "try_transparent: destination already set");
		return CMD_HANDLED;
	}

	t_destination = socketinfo_get_transparent_target(cs);
	/* assume no error - because child_setup() ran */
	ASSUME(!t_destination.error);
	/* ci->destination was not set, so no free() necessary */
	ci->destination = address_ip_mchar(t_destination);
	ci->destinationport = address_port_hs(t_destination);
	/* XXX maybe there was no port set... Call apply_defaults somehow */
	apply_missing_login_values(pi);
	address_free(t_destination);

	jlog(8, "Using transparent proxy. Connecting to %s port %d. User: %s",
						ci->destination,
						ci->destinationport,
						ci->user);

	return CMD_HANDLED;
}


static
int try_transparent_forward(struct proxyinfo* pi,
				struct messagebuffer* msgbuf, int stage) {

	int cs = pi->si->connection.clientsocket;
	struct clientinfo* ci = pi->ci;
	struct serverinfo* si = pi->si;

	const char* transforward_opt = config_get_option("transparent-forward");
	struct address_t destination, t_destination;
	char* transparent_target;
	char* new_user;

	if (si->connection.transparent == TRANSPARENT_NO) {
		return CMD_HANDLED;
	}

	if (!transforward_opt) {
		return CMD_HANDLED;
	}

	/* set the target for the transparent proxy */
	if (!ci->destination) {
		/* replace dest by the specified forward */
		destination = address_parse_host_c_port(transforward_opt);
		if (destination.error) {
			message_buffer_add(msgbuf, cs,
					500, "Error logging in\r\n");
			jlog(4, "Could not parse transparent forward target\r\n");
			return CMD_ERROR;
		}
		free(ci->destination);
		ci->destination = address_host_pref_name_mchar(destination);
		ci->destinationport = address_port(destination);
		address_free(destination);

		jlog(9, "Enabling transparent forward to %s:%d",
				ci->destination, ci->destinationport);
	}

	/* XXX ? */ /*
	if (!ci->forward.original.user && ci->user) {
		char* myself;
		struct in_addr myself_in;

		ci->forward.original.user = strdup(ci->user);
		* if there was a transparent forward, the original
		 * destination was the proxy's interface *
		ci->forward.original.dest_ip
			= socketinfo_get_local_ip(cs);
		myself_in.s_addr = ci->forward.original.dest_ip;
		myself = inet_ntoa(myself_in);
		if (myself) {
			ci->forward.original.destination = strdup(myself);
		} else {
			ci->forward.original.destination = " -error- ";
		}
		ci->forward.original.destinationport
			= socketinfo_get_local_port(cs);
	}*/

	t_destination = socketinfo_get_transparent_target(cs);
	/* assume no error - because child_setup() ran */
	ASSUME(!t_destination.error);
	/* rewrite the user name, append the transparent destination */
	transparent_target = address_ip_c_port_mchar(t_destination);
	if (config_get_bool("transparent-forward-include-port", 0) == 0) {
		char* colon = strrchr(transparent_target, ':');
		if (colon) { *colon = '\0'; }
	}
	/* we do not necessarily have a user. We can have logintime connect
	 * and have a transparent-forward */
	if (ci->user) {
		new_user = char_enclose(ci->user, "@", transparent_target);
		free(ci->user);
		ci->user = new_user;
	}
	address_free(t_destination);
	free(transparent_target);

	jlog(8, "Using transparent forward. Connecting to %s port %d. User: %s",
						ci->destination,
						ci->destinationport,
						ci->user);

	return CMD_HANDLED;
}

static
int fw_validate_user_authprog_extended(struct option_t* ctx,
				       const char* prog_name,
				       const char* prog_args[],
				       struct proxyinfo* pi) {

	int ret;

	/* add other variables, read from pi to *ctx */
	ret = auth_validate_external(ctx, AUTH_LEVEL_EXTENDED,
					prog_name, prog_args);
	return ret;
}

static
int fw_validate_user_authprog(const char* fwuser,
			      const char* fwpass,
			      int proglevel,
			      const struct slist_t* proginfo,
			      struct proxyinfo* pi) {

	char* prog_name;
	char** prog_args;
	struct option_t* ctx;
	int i, j, ret;

	if (!proginfo || !proginfo->value) {
		return AUTH_FAIL;
	}

	prog_name = proginfo->value;

	i = slist_count(proginfo->next);
	/*
	 * args[0]: program name
	 * args[1-n]: argument
	 * args[n+1]: terminating 0
	 * => for n arguments, we have n+2 elements in the array */
	prog_args = (char**) malloc((i + 2) * sizeof(char*));

	proginfo = proginfo->next;
	prog_args[0] = prog_name;
	for (j = 1; j < i+1; j++) {
		prog_args[j] = proginfo->value;
		proginfo = proginfo->next;
	}
	prog_args[j] = (char*) 0;

	ctx = optionlist_init("username", fwuser);
	optionlist_push(ctx, "password", fwpass);

	if (proglevel == AUTH_LEVEL_SIMPLE) {
		ret = auth_validate_external(ctx, AUTH_LEVEL_SIMPLE, prog_name,
						(const char**) prog_args);
	} else {
		ret = fw_validate_user_authprog_extended(ctx, prog_name,
						(const char**) prog_args, pi);
	}

	optionlist_destroy(ctx);
	free(prog_args);

	return ret;
}

static
int fw_validate_user(const char* fwuser,
		     const char* fwpass,
		     const struct slist_t* account) {

	const char* user, *pass;
	if (!account || !account->next) {
		return AUTH_FAIL;
	}

	user = account->value;

	if (strcmp(user, fwuser) != 0) {
		return AUTH_FAIL;
	}

	pass = account->next->value;

	if (cryptcmp(pass, fwpass) != 0) {
		jlog(8, "Authentication of user \"%s\" by means of account "
				"setting failed", user);
		return AUTH_FAIL;
	} else {
		jlog(8, "Authentication of user \"%s\" by means of account "
				"setting was ok", user);
	}

	return AUTH_OK;
}

/* returns 0 if the user has supplied the correct password */
static
int fw_validate(const char* fwuser, const char* fwpass, struct proxyinfo* pi) {
	struct slist_t* acct_list = config_get_option_array("account");
	struct slist_t* account, *acct_line;
	const char* method;
	int ret = AUTH_FAIL;

	if ( ! acct_list ) {
		jlog(5, "No account information found for %s", fwuser);
		return AUTH_FAIL;
	}

	/* reverse the list to keep the paradigma: if there are several
	 * similar options, the last one is taken */
	acct_list = slist_reverse(acct_list);

	account = acct_list;

	do {
		acct_line = config_split_line(account->value, WHITESPACES);

		if (!acct_line || !acct_line->value
			|| !acct_line->next || !acct_line->next->value) {

			jlog(5, "Incorrect account specification: %s",
							account->value);
			slist_destroy_free(acct_line);
			continue;
		}
		method = acct_line->value;
		if (strcasecmp(method, "single") == 0) {
			ret = fw_validate_user(fwuser, fwpass, acct_line->next);
		} else if (strcasecmp(method, "authProgramSimple") == 0) {
			ret = fw_validate_user_authprog(fwuser, fwpass,
						AUTH_LEVEL_SIMPLE,
						acct_line->next,
						pi);
		} else if (strcasecmp(method, "authProgramExtended") == 0) {
			ret = fw_validate_user_authprog(fwuser, fwpass,
						AUTH_LEVEL_EXTENDED,
						acct_line->next,
						pi);
		} else {
			/* XXX syntax has changed for 0.14.x */
			ASSUME(method != NULL);
			jlog(5, "Unknown account type \"%s\"", method);
		}
		if (ret == AUTH_OK) {
			slist_destroy_free(acct_line);
			slist_destroy_free(acct_list);
			return AUTH_OK;
		}
		/* AUTH_FAIL */
		continue;
	} while ( (account = account->next) );

	slist_destroy_free(acct_list);

	return AUTH_FAIL;
}

static
void fw_say_welcome(struct proxyinfo* pi, struct messagebuffer* msgbuf) {
	int cs = pi->si->connection.clientsocket;
	const char* welcome = config_get_option("welcomeline");

	if (welcome) {
		/* XXX check */
		message_buffer_add(msgbuf, cs, 220, welcome);
	} else {
		message_buffer_add(msgbuf, cs, 220, 
		  "Joe FTP Proxy Server/Gateway (v"PACKAGE_VERSION") ready\r\n");
	}
}

int std_reset(const char* cmd, const char** args,
			struct proxyinfo* pi, struct messagebuffer* msgbuf) {
	/* set all values to (char*) 0 after free()ing */
	struct clientinfo* ci = pi->ci;
	struct serverinfo* si = pi->si;

	jlog(8, "resetting login information");

	if (!si->connection.transparent == TRANSPARENT_YES ||
	    !config_compare_option("logintime", "connect")) {

		ci->login.stage = LOGIN_ST_NOT_CONNECTED;
		free_and_zero_char(&ci->destination);
		ci->destinationport = 0;
	}
	free_and_zero_char(&ci->user);
	free_and_zero_char(&ci->user_login);
	free_and_zero_char(&ci->pass);
	free_and_zero_char(&ci->anon_user_pass);
	/* forward substructure */
	ci->forward.active = 0;
	free_and_zero_char(&ci->forward.accept_pw);
	free_and_zero_char(&ci->forward.send_pw);
	free_and_zero_char(&ci->forward.login);
	free_and_zero_char(&ci->forward.original.user);
	free_and_zero_char(&ci->forward.original.destination);
	ci->forward.original.destinationport = 0;
	/* login substructure */
	free_and_zero_char(&ci->login.fw_auth.user);
	free_and_zero_char(&ci->login.fw_auth.pass);
	free_and_zero_char(&ci->login.msg_welcome);
	free_and_zero_char(&ci->login.msg_after_user);
	free_and_zero_char(&ci->login.msg_after_pass);

	/* we have to reset the command handler */
	cmd_handlers_reset();

	return CMD_HANDLED;
}

static
int std_user_split(const char* cmd, const char** args,
			struct proxyinfo* pi, struct messagebuffer* msgbuf) {

	struct ftp_login_t login_values;
	struct clientinfo* ci = pi->ci;
	int ret;

	/* set the verbatim user name used for the login for the passallauth
	 * option */
	pi->ci->user_login = strdup(args[0]);
	enough_mem(pi->ci->user_login);

	ret = parse_login_string(args[0], &login_values);
	if (ret) {
		return CMD_ERROR;
	}
	apply_default_login_values(&login_values);
	set_login_values_if_undefined(pi, &login_values);

	/* XXX add some enough_mem's */
	if (ci->forward.active) {
		if (!ci->forward.original.user) {
			ci->forward.original.user
				= strdup(login_values.user); }
		if (!ci->forward.original.destination) {
			if (login_values.destination) {
				ci->forward.original.destination
					= strdup(login_values.destination);
			} else {
				ci->forward.original.destination
					= strdup("");
			}
		}
		if (!ci->forward.original.destinationport) {
			ci->forward.original.destinationport
				= login_values.destinationport;
		}
	}

	ret = cmds_after_client_user(pi, msgbuf);
	/* we get CMD_DONE if we are logged in completly, i.e. if the server
	 * didn't ask for a password */
	if (ret != CMD_HANDLED && ret != CMD_DONE) {
		return CMD_ERROR;
	}

	jlog(7, "Client logged in: User: %s, Dest: %s:%d",
					ci->user,
					ci->destination,
					ci->destinationport);

	/* ret can be CMD_HANDLED or CMD_DONE */
	return ret;
}

/* this is not called by std_user_split */
int set_remote_user(const char* user, struct proxyinfo* pi,
					struct messagebuffer *msgbuf) {
	int cs = pi->si->connection.clientsocket;
	struct clientinfo* ci = pi->ci;

	if (!user || !user[0]) {
		message_buffer_add(msgbuf, cs,
			530, "Not a valid username\r\n");
		message_buffer_add(msgbuf, cs,
			530, "Login failed\r\n");
		return CMD_ERROR;
	}

	/* if we are in a forward, fill the user field if it is not yet */
	if (ci->forward.active && !ci->forward.original.user) {
		ci->forward.original.user = strdup(user);
		enough_mem(ci->forward.original.user);
	}

	/* note that by a forward the user can already b set!
	 * Imagine:
	 *
	 * logintime connect
	 * <from 1.2.3.4>
	 * 	forward fred@destination
	 * </from>
	 *
	 * In this case we just set the copy to forward.original.user
	 * */
	if (ci->forward.active && ci->user) {
		/* nothing to do */
		return CMD_HANDLED;
	}
	ASSUME(ci->user == NULL);
	ci->user = strdup(user);
	enough_mem(ci->user);
	return CMD_HANDLED;
}

int std_user_plain(const char* cmd, const char** args,
			struct proxyinfo* pi, struct messagebuffer* msgbuf) {
	/* there will be no splitting */
	if (set_remote_user(args[0], pi, msgbuf) != CMD_HANDLED) {
		/* the function will log an error */
		return CMD_ERROR;
	}

	pi->ci->user_login = strdup(args[0]);
	enough_mem(pi->ci->user_login);

	/* ret can be CMD_HANDLED or CMD_DONE */
	return cmds_after_client_user(pi, msgbuf);
}

int remote_open(const char* cmd, const char** args,
			struct proxyinfo* pi, struct messagebuffer* msgbuf) {
	int cs = pi->si->connection.clientsocket;
	struct ftp_login_t login_values;
	int ret;

	/* the commandhandler wouldn't pass the command here */
	ASSUME(args[0] != NULL);

	ret = parse_login_string_without_user(args[0], &login_values);
	/* if the argument contains an @-sign, the parsing routine did the
	 * wrong thing... */
	if (strchr(args[0], '@') || ret) {
		message_buffer_add(msgbuf, cs,
			530, "Not a valid destination\r\n");
		message_buffer_add(msgbuf, cs,
			530, "Login failed\r\n");
		return CMD_ERROR;
	}

	set_login_values_if_undefined(pi, &login_values);
	apply_missing_login_values(pi);

	message_buffer_add(msgbuf, cs,
			220, "Welcome. Please proceed.\r\n");
	return CMD_HANDLED;
}

int remote_site(const char* cmd, const char** args,
			struct proxyinfo* pi, struct messagebuffer* msgbuf) {
	/* the function will log an error */
	return remote_open(cmd, args, pi, msgbuf);
}

static
int set_remote_pass(const char* pass, struct proxyinfo* pi,
					struct messagebuffer *msgbuf) {
	struct clientinfo* ci = pi->ci;

	/* just chop off the command and put everything in the password */
	if (!pass) {
		/* allow empty passwords, too */
		ci->pass = strdup("");
	} else {
		ci->pass = strdup(pass);
	}
	enough_mem(ci->pass);
	/* at the moment this function returns always success
	 * (this comment included in other places here as well in this file) */
	return CMD_HANDLED;
}

int remote_pass(const char* cmd, const char** args,
			struct proxyinfo* pi, struct messagebuffer* msgbuf) {
	if (set_remote_pass(args[0], pi, msgbuf) != CMD_HANDLED) {
		/* at the moment this function returns always success */
		return CMD_ERROR;
	}
	return cmds_after_client_pass(pi, msgbuf);
}

static
int set_auth_pass(const char* pass, struct proxyinfo* pi,
					struct messagebuffer* msgbuf) {
	int cs = pi->si->connection.clientsocket;
	struct clientinfo* ci = pi->ci;

	/* XXX Can this happen at all ? */
	if (!pass) {
		message_buffer_add(msgbuf, cs,
			530, "Not a valid password\r\n");
		message_buffer_add(msgbuf, cs,
			530, "Login failed\r\n");
		return CMD_ERROR;
	}
	ci->login.fw_auth.pass = strdup(pass);
	enough_mem(ci->login.fw_auth.pass);
	return CMD_HANDLED;
}

int auth_pass(const char* cmd, const char** args,
			struct proxyinfo* pi, struct messagebuffer* msgbuf) {
	if (set_auth_pass(args[0], pi, msgbuf) != CMD_HANDLED) {
		/* the function will log an error */
		return CMD_ERROR;
	}
	/* the function will log an error */
	return cmds_after_auth_pass(pi, msgbuf);
}

static
int set_auth_user(const char* user, struct proxyinfo* pi,
					struct messagebuffer *msgbuf) {
	int cs = pi->si->connection.clientsocket;
	struct clientinfo* ci = pi->ci;

	if (!user || !user[0]) {
		message_buffer_add(msgbuf, cs,
			530, "Not a valid user name\r\n");
		message_buffer_add(msgbuf, cs,
			530, "Login failed\r\n");
		return CMD_ERROR;
	}
	ci->login.fw_auth.user = strdup(user);
	enough_mem(ci->login.fw_auth.user);
	return CMD_HANDLED;
}

int auth_user(const char* cmd, const char** args,
			struct proxyinfo* pi, struct messagebuffer* msgbuf) {
	int cs = pi->si->connection.clientsocket;

	if (set_auth_user(args[0], pi, msgbuf) != CMD_HANDLED) {
		return CMD_ERROR;
	}
	message_buffer_add(msgbuf, cs,
			331, "User name okay, send password.\r\n");
	/* at the moment this function returns always success */
	return cmds_after_auth_user(pi, msgbuf);
}


int fw_user_type8(const char* cmd, const char** args,
			struct proxyinfo* pi, struct messagebuffer* msgbuf) {
	int cs = pi->si->connection.clientsocket;
	struct clientinfo* ci = pi->ci;
	int ret;
	/* expecting "USER fwuser@real.host.name" */

	struct ftp_login_t login_values;

	/* XXX see XXX 1 */
	if ((ret = is_logintime_too_early())) {
		/* 550 because this is a failed USER command */
		message_buffer_add(msgbuf, cs,
			550, "Internal configuration error\r\n");
		return ret;
	}

	/* we call our parsing routine for simplicity though we are dealing
	 * with fwuser, not user.
	 *
	 * Note that by that "fwuser@real.host.name:port,a" is also valid, i.e.
	 * you can specify a mode flag. However, it is ignored, so... */
	ret = parse_login_string(args[0], &login_values);
	if (ret) {
		message_buffer_add(msgbuf, cs,
			530, "Not a valid user name\r\n");
		message_buffer_add(msgbuf, cs,
			530, "Login failed\r\n");
		return CMD_ERROR;
	}
	if (set_auth_user(login_values.user, pi, msgbuf) != CMD_HANDLED) {
		free(login_values.user);
		free(login_values.destination);
		return CMD_ERROR;
	}
	free(login_values.user);

	if (!login_values.destination) {
		message_buffer_add(msgbuf, cs,
			530, "Not a valid user name\r\n");
		message_buffer_add(msgbuf, cs,
			530, "Login failed\r\n");
		return CMD_ERROR;
	}
	ci->destination = login_values.destination;
	ci->destinationport = login_values.destinationport;
	apply_missing_login_values(pi);

	if (cmds_after_auth_user(pi, msgbuf) != CMD_HANDLED) {
		/* at the moment this function returns always success */
		return CMD_ERROR;
	}

	message_buffer_add(msgbuf, cs,
			220, "Welcome. Please proceed.\r\n");

	return CMD_HANDLED;
}

int fw_user_type7(const char* cmd, const char** args,
			struct proxyinfo* pi, struct messagebuffer* msgbuf) {
/*           "USER user@fwuser@real.host.name" and                        */
	int cs = pi->si->connection.clientsocket;

	char* remote_user, *auth_user, *destchar;
	int offset = 0;
	int ret;
	struct ftp_login_t login_values;

	/* XXX - see above */
	/* here we get the password for the firewall account along with the
	 * remote password. So we deny connection until we have verified the
	 * firewall password, i.e. until we got the remote password */
	/* XXX see XXX 1 */
	if ((ret = is_logintime_too_early())) {
		/* 550 because this is a failed USER command */
		message_buffer_add(msgbuf, cs,
			550, "Internal configuration error\r\n");
		return ret;
	}

	remote_user = quotstrtok(args[0], "@", &offset);
	if (set_remote_user(remote_user, pi, msgbuf) != CMD_HANDLED) {
		/* the function will log an error */
		free(remote_user);
		return CMD_ERROR;
	}
	free(remote_user);

	auth_user = quotstrtok(args[0], "@", &offset);
	if (set_auth_user(auth_user, pi, msgbuf) != CMD_HANDLED) {
		free(auth_user);
		return CMD_ERROR;
	}
	free(auth_user);

	destchar = quotstrtok(args[0], "@\n", &offset);
	ret = parse_login_string_without_user(destchar, &login_values);
	free(destchar);
	if (ret) {
		/* XXX 550 ? */
		message_buffer_add(msgbuf, cs,
			550, "Not a valid user name\r\n");
		message_buffer_add(msgbuf, cs,
			550, "Login failed\r\n");
		return CMD_ERROR;
	}
	set_login_values_if_undefined(pi, &login_values);
	apply_missing_login_values(pi);

	if (cmds_after_auth_user(pi, msgbuf) != CMD_HANDLED) {
		/* at the moment this function returns always success */
		return CMD_ERROR;
	}
	/* can be CMD_HANDLED or CMD_DONE */
	return cmds_after_client_user(pi, msgbuf);
}

int fw_pass_type7(const char* cmd, const char** args,
			struct proxyinfo* pi, struct messagebuffer* msgbuf) {
	char* auth_pass, *remote_pass;
	int offset;

	ASSUME(args[0] != NULL);
	
/*      "PASS pass@fwpass" to login.                                 */

	offset = 0;
	/* allow empty passwords */
	if (args[0][0] == '@') {
		/* remote password is empty */
		remote_pass = strdup("");
		enough_mem(remote_pass);
		offset++; /* not really necessary... */
	} else {
		remote_pass = quotstrtok(args[0], "@", &offset);
	}
	if (set_remote_pass(remote_pass, pi, msgbuf) != CMD_HANDLED) {
		/* at the moment this function returns always success */
		free(remote_pass);
		return CMD_ERROR;
	}
	free(remote_pass);

	auth_pass = quotstrtok(args[0], "@\n", &offset);
	if (set_auth_pass(auth_pass, pi, msgbuf) != CMD_HANDLED) {
		/* the function will log an error */
		free(auth_pass);
		return CMD_ERROR;
	}
	free(auth_pass);

	/* XXX we have to merge both responses... */
	if (cmds_after_auth_pass(pi, msgbuf) != CMD_HANDLED) {
		/* the function will log an error */
		return CMD_ERROR;
	}
	return cmds_after_client_pass(pi, msgbuf);
}

/* Here we have two arguments */
int fw_user_type9(const char* cmd, const char** args,
			struct proxyinfo* pi, struct messagebuffer* msgbuf) {
/*           "USER user@real.host.name fwuser"        */
	int cs = pi->si->connection.clientsocket;
	const char* auth_user = args[1];

	if (set_auth_user(auth_user, pi, msgbuf) != CMD_HANDLED) {
		/* the function will log an error */
		return CMD_ERROR;
	}

	if (cmds_after_auth_user(pi, msgbuf) != CMD_HANDLED) {
		/* at the moment this function returns always success */
		return CMD_ERROR;
	}

	/* std_user_split won't notice args[1] */
	if (std_user_split(cmd, args, pi, msgbuf) != CMD_HANDLED) {
		message_buffer_add(msgbuf, cs,
			530, "Not a valid user name\r\n");
		message_buffer_add(msgbuf, cs,
			530, "Login failed\r\n");
		return CMD_ERROR;
	}

	return CMD_HANDLED;
}


int fw_pass_type9(const char* cmd, const char** args,
			struct proxyinfo* pi, struct messagebuffer* msgbuf) {
	/* Just register the password */
	int cs = pi->si->connection.clientsocket;

	/* we allow empty passwords */
	if (set_remote_pass(args[0], pi, msgbuf) != CMD_HANDLED) {
		/* at the moment this function returns always success */
		return CMD_ERROR;
	}
	/* would call cmds_after_client_pass normally but here we postpone it
	 * until after ACCT */
	message_buffer_add(msgbuf, cs,
			332, "Need account for login.\r\n");
	return CMD_HANDLED;
}

int fw_acct_type9(const char* cmd, const char** args,
			struct proxyinfo* pi, struct messagebuffer* msgbuf) {
	if (set_auth_pass(args[0], pi, msgbuf) != CMD_HANDLED) {
		/* the function will log an error */
		return CMD_ERROR;
	}
	if (cmds_after_auth_pass(pi, msgbuf) != CMD_HANDLED) {
		/* the function will log an error */
		return CMD_ERROR;
	}
	return cmds_after_client_pass(pi, msgbuf);
}

/* Some of the following firewall types are taken out of ncftp's
 * configuration file.  See ftp://ftp.ncftp.com/ncftp by
 *
 * Mike Gleason
 * NcFTP Software
 * mgleason@NcFTP.com
 */

/* type 0:  Do NOT use a firewall                                         */
int login_register_logintype_0 (void) {
	cmd_register(CMD_HANDLERTYPE_IMPLEMENT,
			CMD_SERIALIZE_REMOVE_ONE,
			"QUIT", "standard QUIT", 0, 0,
			-1,
			std_quit,
			(char*) 0, (cmd_hookfunc_t) 0);

	cmd_register(CMD_HANDLERTYPE_IMPLEMENT,
			CMD_SERIALIZE_REMOVE_ONE,
			"USER", "plain USER", 1, 1,
			-1,
			std_user_plain,
			"PASS", std_reset);

	cmd_register(CMD_HANDLERTYPE_IMPLEMENT,
			CMD_SERIALIZE_REMOVE_ONE,
			"PASS", "standard PASS", 0, 1,
			-1,
			remote_pass,
			(char*) 0, (cmd_hookfunc_t) 0);

	cmd_set_initial_expect("USER");
	return 0;
}


/* type 1:  Connect to firewall host, but send "USER user@real.host.name" */
/* "USER with no login (user@host port)"                                  */
/* "USER with no login (user@host:port)"                                  */
/* "USER with no login (user@host:port,a)"                                */
/* "USER with no login (user@host:port,p)"                                */
/* "USER with no login ("user with spaces"@host:port,p)"                  */
/* "USER with no login ("user@otherhost"@host:port,p)"                    */

int login_register_logintype_1 (void) {
	cmd_register(CMD_HANDLERTYPE_IMPLEMENT,
			CMD_SERIALIZE_REMOVE_ONE,
			"QUIT", "standard QUIT", 0, 0,
			-1,
			std_quit,
			(char*) 0, (cmd_hookfunc_t) 0);

	cmd_register(CMD_HANDLERTYPE_IMPLEMENT,
			CMD_SERIALIZE_REMOVE_ONE,
			"USER", "splitting USER", 1, 1,
			-1,
			std_user_split,
			"PASS", std_reset);

	/* allow empty passwords, too */
	cmd_register(CMD_HANDLERTYPE_IMPLEMENT,
			CMD_SERIALIZE_REMOVE_ONE,
			"PASS", "standard PASS", 0, 1,
			-1,
			remote_pass,
			(char*) 0, std_reset);

	cmd_set_initial_expect("USER");
	return 0;
}

/* type 2:  Connect to firewall, login with "USER fwuser" and             */
/*          "PASS fwpassword", and then "USER user@real.host.name"        */
/* joe:     and thereafter "PASS password"                                */
/* "USER with login (user@host port)"                                  */
/* "USER with login (user@host:port)"                                  */
int login_register_logintype_2 (void) {
	cmd_register(CMD_HANDLERTYPE_IMPLEMENT,
			CMD_SERIALIZE_REMOVE_ONE,
			"QUIT", "standard QUIT", 0, 0,
			-1,
			std_quit,
			(char*) 0, (cmd_hookfunc_t) 0);

	cmd_register(CMD_HANDLERTYPE_IMPLEMENT,
			CMD_SERIALIZE_REMOVE_ONE,
			"USER", "authenticating USER", 1, 1,
			0,
			auth_user,
			"PASS", std_reset);

	/* don't allow empty authenticating passwords */
	cmd_register(CMD_HANDLERTYPE_IMPLEMENT,
			CMD_SERIALIZE_REMOVE_ONE,
			"PASS", "authenticating PASS", 1, 1,
			0,
			auth_pass,
			"USER", std_reset);

	cmd_register(CMD_HANDLERTYPE_IMPLEMENT,
			CMD_SERIALIZE_REMOVE_ONE,
			"USER", "splitting USER", 1, 1,
			1,
			std_user_split,
			"PASS", std_reset);

	/* allow empty passwords, too */
	cmd_register(CMD_HANDLERTYPE_IMPLEMENT,
			CMD_SERIALIZE_REMOVE_ONE,
			"PASS", "standard PASS", 0, 1,
			1,
			remote_pass,
			(char*) 0, std_reset);

	cmd_set_initial_expect("USER");
	return 0;
}

/* type 3:  Connect to and login to firewall, and then use                */
/*          "SITE real.host.name", followed by the regular USER and PASS. */
/* "SITE with login (user@host port)"                                  */
/* "SITE with login (user@host:port)"                                  */

int login_register_logintype_3 (void) {
	cmd_register(CMD_HANDLERTYPE_IMPLEMENT,
			CMD_SERIALIZE_REMOVE_ONE,
			"QUIT", "standard QUIT", 0, 0,
			-1,
			std_quit,
			(char*) 0, (cmd_hookfunc_t) 0);

	cmd_register(CMD_HANDLERTYPE_IMPLEMENT,
			CMD_SERIALIZE_REMOVE_ONE,
			"USER", "authenticating USER", 1, 1,
			0,
			auth_user,
			"PASS", std_reset);

	/* don't allow empty authenticating passwords */
	cmd_register(CMD_HANDLERTYPE_IMPLEMENT,
			CMD_SERIALIZE_REMOVE_ONE,
			"PASS", "authenticating PASS", 1, 1,
			0,
			auth_pass,
			"SITE", std_reset);

	cmd_register(CMD_HANDLERTYPE_IMPLEMENT,
			CMD_SERIALIZE_REMOVE_ONE,
			"SITE", "destination SITE", 1, 1,
			-1,
			remote_site,
			"USER", std_reset);

	cmd_register(CMD_HANDLERTYPE_IMPLEMENT,
			CMD_SERIALIZE_REMOVE_ONE,
			"USER", "destination USER", 1, 1,
			1,
			std_user_plain,
			"PASS", std_reset);

	/* allow empty passwords, too */
	cmd_register(CMD_HANDLERTYPE_IMPLEMENT,
			CMD_SERIALIZE_REMOVE_ONE,
			"PASS", "destination PASS", 0, 1,
			1,
			remote_pass,
			(char*) 0, std_reset);

	cmd_set_initial_expect("USER");
	return 0;
}

/* type 4:  Connect to firewall, and then use (without login)             */
/*          "SITE real.host.name", followed by the regular USER and PASS. */
/*  SITE without login                                                    */

int login_register_logintype_4 (void) {
	cmd_register(CMD_HANDLERTYPE_IMPLEMENT,
			CMD_SERIALIZE_REMOVE_ONE,
			"QUIT", "standard QUIT", 0, 0,
			-1,
			std_quit,
			(char*) 0, (cmd_hookfunc_t) 0);

	cmd_register(CMD_HANDLERTYPE_IMPLEMENT,
			CMD_SERIALIZE_REMOVE_ONE,
			"SITE", "destination SITE", 1, 1,
			-1,
			remote_site,
			"USER", std_reset);

	cmd_register(CMD_HANDLERTYPE_IMPLEMENT,
			CMD_SERIALIZE_REMOVE_ONE,
			"USER", "destination USER", 1, 1,
			-1,
			std_user_plain,
			"PASS", std_reset);

	/* allow empty passwords, too */
	cmd_register(CMD_HANDLERTYPE_IMPLEMENT,
			CMD_SERIALIZE_REMOVE_ONE,
			"PASS", "destination PASS", 0, 1,
			-1,
			remote_pass,
			(char*) 0, std_reset);

	cmd_set_initial_expect("SITE");
	return 0;
}

/* type 5:  Connect to and login to firewall, and then use                */
/*          "OPEN real.host.name", followed by the regular USER and PASS. */

int login_register_logintype_5 (void) {
	cmd_register(CMD_HANDLERTYPE_IMPLEMENT,
			CMD_SERIALIZE_REMOVE_ONE,
			"QUIT", "standard QUIT", 0, 0,
			-1,
			std_quit,
			(char*) 0, (cmd_hookfunc_t) 0);

	cmd_register(CMD_HANDLERTYPE_IMPLEMENT,
			CMD_SERIALIZE_REMOVE_ONE,
			"USER", "authenticating USER", 1, 1,
			0,
			auth_user,
			"PASS", std_reset);

	/* don't allow empty authenticating passwords */
	cmd_register(CMD_HANDLERTYPE_IMPLEMENT,
			CMD_SERIALIZE_REMOVE_ONE,
			"PASS", "authenticating PASS", 1, 1,
			0,
			auth_pass,
			"OPEN", std_reset);

	cmd_register(CMD_HANDLERTYPE_IMPLEMENT,
			CMD_SERIALIZE_REMOVE_ONE,
			"OPEN", "destination OPEN", 1, 1,
			-1,
			remote_open,
			"USER", std_reset);

	cmd_register(CMD_HANDLERTYPE_IMPLEMENT,
			CMD_SERIALIZE_REMOVE_ONE,
			"USER", "destination USER", 1, 1,
			1,
			std_user_plain,
			"PASS", std_reset);

	/* allow empty passwords, too */
	cmd_register(CMD_HANDLERTYPE_IMPLEMENT,
			CMD_SERIALIZE_REMOVE_ONE,
			"PASS", "destination PASS", 0, 1,
			1,
			remote_pass,
			(char*) 0, std_reset);

	cmd_set_initial_expect("USER");
	return 0;
}

/* type 6:  Connect to firewall, and then use (without login)             */
/*          "OPEN real.host.name", followed by the regular USER and PASS. */


int login_register_logintype_6 (void) {
	cmd_register(CMD_HANDLERTYPE_IMPLEMENT,
			CMD_SERIALIZE_REMOVE_ONE,
			"QUIT", "standard QUIT", 0, 0,
			-1,
			std_quit,
			(char*) 0, (cmd_hookfunc_t) 0);

	cmd_register(CMD_HANDLERTYPE_IMPLEMENT,
			CMD_SERIALIZE_REMOVE_ONE,
			"OPEN", "destination OPEN", 1, 1,
			-1,
			remote_site,
			"USER", std_reset);

	cmd_register(CMD_HANDLERTYPE_IMPLEMENT,
			CMD_SERIALIZE_REMOVE_ONE,
			"USER", "destination USER", 1, 1,
			-1,
			/* fw_user */
			std_user_plain,
			"PASS", std_reset);

	/* allow empty passwords, too */
	cmd_register(CMD_HANDLERTYPE_IMPLEMENT,
			CMD_SERIALIZE_REMOVE_ONE,
			"PASS", "destination PASS", 0, 1,
			-1,
			remote_pass,
			(char*) 0, std_reset);

	cmd_set_initial_expect("OPEN");
	return 0;
}

/* type 7:  Connect to firewall host, but send                            */
/*           "USER user@fwuser@real.host.name" and                        */
/*           "PASS pass@fwpass" to login.                                 */
/* CheckPoint Firewall 1                                                  */

int login_register_logintype_7 (void) {
	cmd_register(CMD_HANDLERTYPE_IMPLEMENT,
			CMD_SERIALIZE_REMOVE_ONE,
			"QUIT", "standard QUIT", 0, 0,
			-1,
			std_quit,
			(char*) 0, (cmd_hookfunc_t) 0);

	cmd_register(CMD_HANDLERTYPE_IMPLEMENT,
			CMD_SERIALIZE_REMOVE_ONE,
			"USER", "loginstyle 7 USER", 1, 1,
			-1,
			fw_user_type7,
			"PASS", std_reset);

	cmd_register(CMD_HANDLERTYPE_IMPLEMENT,
			CMD_SERIALIZE_REMOVE_ONE,
			"PASS", "loginstyle 7 PASS", 1, 1,
			-1,
			fw_pass_type7,
			(char*) 0, std_reset);

	cmd_set_initial_expect("USER");
	return 0;
}

/*  type 8:  Connect to firewall host, but send                           */
/*           "USER fwuser@real.host.name" and                             */
/*           "PASS fwpass" followed by a regular                          */
/*           "USER user" and                                              */
/*           "PASS pass" to complete the login.                           */

int login_register_logintype_8 (void) {
	cmd_register(CMD_HANDLERTYPE_IMPLEMENT,
			CMD_SERIALIZE_REMOVE_ONE,
			"QUIT", "standard QUIT", 0, 0,
			-1,
			std_quit,
			(char*) 0, (cmd_hookfunc_t) 0);

	cmd_register(CMD_HANDLERTYPE_IMPLEMENT,
			CMD_SERIALIZE_REMOVE_ONE,
			"USER", "loginstyle 8 USER", 1, 1,
			0,
			fw_user_type8,
			"PASS", std_reset);

	/* don't allow empty authenticating passwords */
	cmd_register(CMD_HANDLERTYPE_IMPLEMENT,
			CMD_SERIALIZE_REMOVE_ONE,
			"PASS", "authenticating PASS", 1, 1,
			0,
			auth_pass,
			"USER", std_reset);

	cmd_register(CMD_HANDLERTYPE_IMPLEMENT,
			CMD_SERIALIZE_REMOVE_ONE,
			"USER", "standard USER", 1, 1,
			1,
			std_user_plain,
			"PASS", std_reset);

	/* allow empty passwords, too */
	cmd_register(CMD_HANDLERTYPE_IMPLEMENT,
			CMD_SERIALIZE_REMOVE_ONE,
			"PASS", "standard PASS", 0, 1,
			1,
			remote_pass,
			(char*) 0, std_reset);

	cmd_set_initial_expect("USER");
	return 0;
}

/*  type 9:  Connect to firewall host, but send                           */
/*           "USER user@real.host.name fwuser" and                        */
/*           "PASS pass" followed by                                      */
/*           "ACCT fwpass" to complete the login.                         */
/* User@host FireID                                                       */

int login_register_logintype_9 (void) {
	cmd_register(CMD_HANDLERTYPE_IMPLEMENT,
			CMD_SERIALIZE_REMOVE_ONE,
			"QUIT", "standard QUIT", 0, 0,
			-1,
			std_quit,
			(char*) 0, (cmd_hookfunc_t) 0);

	cmd_register(CMD_HANDLERTYPE_IMPLEMENT,
			CMD_SERIALIZE_REMOVE_ONE,
			/* Yeah! 2 arguments! */
			"USER", "loginstyle 9 USER", 2, 2,
			0,
			fw_user_type9,
			"PASS", std_reset);

	/* allow empty passwords, too */
	cmd_register(CMD_HANDLERTYPE_IMPLEMENT,
			CMD_SERIALIZE_REMOVE_ONE,
			"PASS", "loginstyle 9 PASS", 0, 1,
			0,
			/* normal pass routine but it answers with "Need
			 * account for login" */
			fw_pass_type9,
			"ACCT", std_reset);

	cmd_register(CMD_HANDLERTYPE_IMPLEMENT,
			CMD_SERIALIZE_REMOVE_ONE,
			"ACCT", "loginstyle 9 ACCT", 1, 1,
			0,
			fw_acct_type9,
			(char*) 0, std_reset);

	cmd_set_initial_expect("USER");
	return 0;
}

int fw_install_handler(int number) {
	cmd_handlers_reset();
	switch(number) {
		case 0: return login_register_logintype_0();
		case 1: return login_register_logintype_1();
		case 2: return login_register_logintype_2();
		case 3: return login_register_logintype_3();
		case 4: return login_register_logintype_4();
		case 5: return login_register_logintype_5();
		case 6: return login_register_logintype_6();
		case 7: return login_register_logintype_7();
		case 8: return login_register_logintype_8();
		case 9: return login_register_logintype_9();
	}
	return -1;
}

