/* 
 * Copyright (C) 1999-2004 Joachim Wieland <joe@mcknight.de>
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place - Suite 330, Boston, MA 02111, USA.
 */

#include <signal.h>
#include <ctype.h>
#include <fcntl.h>
#include "jproxygw.h"

#define MAXSIZE 200

/* reads a single line from fd and returns a pointer to a malloc()ed 
 * char array.
 *
 * Parameters: fd: The file descriptor to read from
 *
 * Return values: (char*) 0 on error, a pointer to a malloc()ed char array
 *                that contains the read data on success
 *
 * Called by: various functions
 */

extern int timeout;

int got_sig_pipe;


char* readline(int fd) {
	/* do not check for valid FTP responses */
	int tmo = config_get_ioption("commandtimeout", 300);
	return readline_check_timeout(fd, 0, tmo);
}

char* readline_timeout(int fd, int timeout_sec) {
	return readline_check_timeout(fd, 0, timeout_sec);
}

char *readline_check_timeout(int fd, int check_ftp_format, int timeout_sec) {
	/* XXX increase before release */
	const size_t MAXCHUNKSIZE = 3;
	int n, ret;
	int linecnt;
	int data_read = 0;
	int maxlength_exceeded = 0;
	size_t length = 0;
	size_t real_length = 0;
	char *temp, *buffer;
	fd_set readset;
	struct timeval rtimeout;
	sigset_t sigset, oldset;
	rtimeout.tv_sec = timeout_sec;
	rtimeout.tv_usec = 0;

	FD_ZERO(&readset);
	FD_SET(fd, &readset);

	timeout = 0;

	sigemptyset(&sigset);
	sigemptyset(&oldset);
	sigaddset(&sigset, SIGCHLD);
	ret = sigprocmask(SIG_BLOCK, &sigset, &oldset);
	if (ret < 0) {
		jlog(3, "sigprocmask() error: %s", strerror(errno));
	}
	if ((ret = select(fd + 1, &readset, NULL, NULL, &rtimeout)) <= 0) {
		if (ret == 0) {
			timeout = 1;
		}
		/* save the errno value from sigprocmask() */
		n = errno;
		ret = sigprocmask(SIG_UNBLOCK, &sigset, &oldset);
		if (ret < 0) {
			jlog(3, "sigprocmask() error releasing the blocked"
				" signals: %s", strerror(errno));
		}
		errno = n;
		return 0;
	}
	ret = sigprocmask(SIG_UNBLOCK, &sigset, &oldset);
	if (ret < 0) {
		jlog(3, "sigprocmask() error releasing the blocked"
			" signals: %s", strerror(errno));
	}
	
	if (fd < 0) {
		jlog(1, "readline_check: Not connected");
		return 0;
	}

	buffer = (char *)malloc(MAXCHUNKSIZE);
	enough_mem(buffer);

	temp = buffer;

	linecnt = 0;
	while ((n = read(fd, temp, 1)) > 0) {
		data_read = 1;
		if (*temp == '\r') { linecnt++; continue; }
		if (*temp == '\n') { linecnt++; break; }
		if (*temp == '\0') break;
		/* check for a valid FTP response. It must start with either
		 * xxx <text> (mind the space)
		 * or
		 * xxx-<text> (with a dash)
		 */
		if (length < MAXSIZE) {
			length++;
			if (check_ftp_format && linecnt == 0
			 	&& ((length <= 3 && !isdigit((int) *temp))
				|| (length == 4 && *temp != ' '
						&& *temp != '-'))) {

				buffer[length] = 0;
				jlog(4, "malformed FTP response: %s",
					buffer);
				jlog(9, "in line: %d", linecnt);
				set_errstr("malformed FTP response");
				return 0;
			}
			if ((length+1) % MAXCHUNKSIZE == 0) {
				buffer = (char*) realloc(buffer, length + 1 + MAXCHUNKSIZE);
				enough_mem(buffer);
				temp = buffer + length - 1;
			}
			temp++;
			real_length = length;
		} else {
			maxlength_exceeded = 1;
			real_length++;
		}
	}
	if (maxlength_exceeded) {
		jlog(4, "Input exceeded maximum length. Returning the first %d bytes of %d bytes read totally", length, real_length);
	}
	if (n < 0 || (n == 0 && data_read == 0)) {
		if (n < 0) {
			set_errstr(strerror(errno));
			jlog(2, "Error reading: %s", strerror(errno));
		}
		free(buffer);
		return 0;
	}
	buffer[length] = '\0';
	/* XXX only for FTP ? */
	if (my_strcasestr(buffer, "PASS ") == (char*) 0) {
		jlog(9, "Read (%d): %s", fd, buffer);
	} else {
		jlog(9, "Read (%d): ***hidden***", fd);
	}
	return buffer;
}


/* writes a char* to an fd and checks the return value */

int say(int fd, const char* phrase) {
	int i;
	struct timeval writetime = { 300, 0 };
	fd_set writeset;
	FD_ZERO(&writeset);
	FD_SET(fd, &writeset);

	i = select(fd + 1, NULL, &writeset, NULL, &writetime);
	if (i == 0) {
		jlog(2, "Timeout reached in say()");
		jlog(2, "Timeout in %s line %d\n", __FILE__ ,__LINE__);
		timeout = 1;
		return -1;
	}
	if (i == -1) {
		jlog(1, "Error in select() in say(): %s", strerror(errno));
		return -1;
	}

	/* XXX only for FTP ? */
	if (my_strcasestr(phrase, "PASS ") == (char*) 0) {
		jlog(9, "Write(%d): %s", fd, phrase);
	} else {
		jlog(9, "Write(%d): ***hidden***", fd);
	}
	i = write(fd, phrase, strlen(phrase));
	if (i < 0) {
		jlog(3, "write (say) failed: %s", strerror(errno));
		if (my_strcasestr(phrase, "PASS ") == (char*) 0) {
			jlog(3, "should say %s to %d", phrase, fd);
		} else {
			jlog(3, "should say ***hidden*** to %d", fd);
		}
	}
	return i;
}

#define SAYBUFFERSIZE		200
int sayf(int fd, const char* fmt, ...) {
	va_list args;
	static char str[SAYBUFFERSIZE];

	va_start(args, fmt);
	vsnprintf(str, SAYBUFFERSIZE - 1, fmt, args);
	va_end(args);

	return say(fd, str);
}

void err_time_readline(int fd) {
	jlog(2, "Timeout reached in readline()");
	say(fd, "500 Connection timed out\r\n");
}

void err_readline(int fd) {
	char* m;
	int e = errno;
	size_t msize;
	const char* err;
	char* s = "An error occurred in ftp_readline: %s";

	if (!(err = get_errstr())) {
		err = strerror(e);
	}
	jlog(2, s, err);
	msize = strlen(s) + strlen(err) + 7;
	m = (char*) malloc(msize);
	enough_mem(m);
	strcpy(m, "500 ");
	snprintf(m + strlen(m),
			msize         /* initial size */
			- strlen(m)   /* "500 " */
			- 3,          /* \r\n\0 */
			s, err);
	strcat(m, "\r\n");
	say(fd, m);
	free(m);
}


/* the difference to the original read(2) function is that this function
 * reads as long as it gets size bytes. If less than size bytes are sent, it
 * hangs...
 * if there is an error, it returns of course */
ssize_t readall(int fd, char* buffer, size_t* size) {
	return readall_checkfunc(fd, buffer, size,
					(struct proxyinfo*) 0,
					(socketio_checkfunc_t) 0);
}

ssize_t readall_checkfunc(int fd, char* buffer, size_t* size,
					struct proxyinfo* pi,
					socketio_checkfunc_t checker_f) {
	ssize_t bytes_read;
	size_t remain = *size;
	size_t offset = 0;
	/* simulate read() behavior. */
	static int finished;

	*size = 0;
	if (finished) {
		finished = 0;
		return 0;
	}
	if (remain == 0) {
		return 0;
	}
	do {
		bytes_read = read(fd, buffer + offset, remain);
		if (bytes_read < 0) {
			jlog(2, "read() error in readall: %s", strerror(errno));
			return -1;
		} else if (bytes_read == 0) {
			/* if we haven't read any data yet, return 0 at once
			 * and don't set the finished-flag */
			if (offset == 0) {
				return 0;
			} else {
				finished = 1;
			}
		} else {
			remain -= bytes_read;
			offset += bytes_read;
			if (checker_f) {
				int ret;
				struct transfer_t transferred;
				transferred.buffer = buffer;
				transferred.bufsize = (*size - remain);
				transferred.status = TRANSFER_STATUS_NORMAL;
				if ((ret = checker_f(&transferred, pi))
						!= TRANSFER_STATUS_NORMAL) {
					/* we have to return a negative value
					 * to attract attention... */
					return -ret;
				}
			}
		}
		*size = offset;
	} while (remain > 0 && bytes_read > 0 && !finished);
	return offset;
}

static
void handler_sig_pipe(int __unused__) {
	got_sig_pipe = 1;
}

static
struct sigaction register_catch_sig_pipe(void) {
	struct sigaction sa, old_sa;

	got_sig_pipe = 0;

	sa.sa_handler = handler_sig_pipe;
	sigemptyset (&sa.sa_mask);
#ifndef WIN32
	sa.sa_flags = SA_RESTART;
#endif
	sigaction (SIGPIPE, &sa, &old_sa);
	return old_sa;
}

static
void unregister_catch_sig_pipe(struct sigaction sa) {
	sigaction (SIGPIPE, &sa, 0);
	got_sig_pipe = 0;
}

/* see above for the analogus difference to the original write(2) function */
ssize_t writeall(int fd, char* buffer, size_t size) {
	return writeall_checkfunc(fd, buffer, size,
					(struct proxyinfo*) 0,
					(socketio_checkfunc_t) 0);
}

ssize_t writeall_checkfunc(int fd, char* buffer, size_t size,
					struct proxyinfo* pi,
					socketio_checkfunc_t checker_f) {
	ssize_t bytes_written;
	size_t offset = 0;
	fd_set writeset;
	struct timeval wtimeout = { 0, 0 };
	struct sigaction sa;
	FD_ZERO(&writeset);
	FD_SET(fd, &writeset);

	/* XXX are we always nonblocking? */

	/* install signal handler - retrieve old behaviour */
	sa = register_catch_sig_pipe();

	/* there is a default value already */
	/* XXX is it always the transfertimeout? Should we make it
	 * parameterized? */
	wtimeout.tv_sec = config_get_ioption("transfertimeout", -1);

	do {
		int i = select(fd + 1, NULL, &writeset, &writeset, &wtimeout);
		if (i < 0) {
			jlog(2, "Error in select while waiting to write: %s",
							strerror(errno));
			unregister_catch_sig_pipe(sa);
			return -1;
		}
		if (i == 0) {
			jlog(6, "Timeout while waiting to write: %s",
							strerror(errno));
			unregister_catch_sig_pipe(sa);
			return -1;
		}
		bytes_written = write(fd, buffer + offset, size);
		if (bytes_written < 0) {
			if (errno == EINTR && got_sig_pipe) {
				jlog(2, "write() error. Remote peer "
						"closed the connection");
				unregister_catch_sig_pipe(sa);
				return -1;
			}
			jlog(2, "write() error in writeall: %s",
					strerror(errno));
			unregister_catch_sig_pipe(sa);
			return -1;
		} else {
			size -= bytes_written;
			offset += bytes_written;
			/* XXX does this make sense at all?? */
			if (checker_f) {
				int ret;
				struct transfer_t transferred;
				transferred.buffer = buffer;
				transferred.bufsize = size;
				transferred.status = TRANSFER_STATUS_NORMAL;
				if ((ret = checker_f(&transferred, pi))
						!= TRANSFER_STATUS_NORMAL) {
					/* we have to return a negative value
					 * to attract attention... */
					return -ret;
				}
			}
		}
	} while (size > 0 && bytes_written >= 0);

	unregister_catch_sig_pipe(sa);
	return offset;
}

int socketio_set_nonblocking(int fd) {
	int fdflags;

	if ((fdflags = fcntl(fd, F_GETFL)) < 0) {
		jlog(2, "Error getting fcntl() flags for fd %d", fd);
		return -1;
	}
	if(fcntl(fd, F_SETFL, fdflags | O_NONBLOCK) < 0) {
		jlog(2, "Error setting fd %d to nonblocking", fd);
		return -1;
	}
	return fdflags;
}

int socketio_set_flags(int fd, int flags) {
	if (fcntl(fd, F_SETFL, flags) < 0) {
		jlog(2, "Error resetting fd flags for fd %d", fd);
		return -1;
	}
	return 0;
}

